from vyper.codegen.ir_node import Encoding, IRnode
from vyper.compiler.settings import _opt_codesize, _opt_gas, _opt_none
from vyper.evm.address_space import (
    CALLDATA,
    DATA,
    IMMUTABLES,
    MEMORY,
    STORAGE,
    TRANSIENT,
    AddrSpace,
)
from vyper.evm.opcodes import version_check
from vyper.exceptions import CompilerPanic, TypeCheckFailure, TypeMismatch
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types import (
    AddressT,
    BoolT,
    BytesM_T,
    BytesT,
    DArrayT,
    DecimalT,
    HashMapT,
    IntegerT,
    InterfaceT,
    StructT,
    TupleT,
    _BytestringT,
)
from vyper.semantics.types.shortcuts import BYTES32_T, INT256_T, UINT256_T
from vyper.semantics.types.subscriptable import SArrayT
from vyper.semantics.types.user import FlagT
from vyper.utils import GAS_COPY_WORD, GAS_IDENTITY, GAS_IDENTITYWORD, ceil32

DYNAMIC_ARRAY_OVERHEAD = 1


def is_bytes_m_type(typ):
    return isinstance(typ, BytesM_T)


def is_numeric_type(typ):
    return isinstance(typ, (IntegerT, DecimalT))


def is_integer_type(typ):
    return isinstance(typ, IntegerT)


def is_decimal_type(typ):
    return isinstance(typ, DecimalT)


def is_flag_type(typ):
    return isinstance(typ, FlagT)


def is_tuple_like(typ):
    # A lot of code paths treat tuples and structs similarly
    # so we have a convenience function to detect it
    ret = isinstance(typ, (TupleT, StructT))
    assert ret == hasattr(typ, "tuple_items")
    return ret


def is_array_like(typ):
    # For convenience static and dynamic arrays share some code paths
    ret = isinstance(typ, (DArrayT, SArrayT))
    assert ret == typ._is_array_type
    return ret


def get_type_for_exact_size(n_bytes):
    """Create a type which will take up exactly n_bytes. Used for allocating internal buffers.

    Parameters:
      n_bytes: the number of bytes to allocate
    Returns:
      type: A type which can be passed to context.new_variable
    """
    return BytesT(n_bytes - 32 * DYNAMIC_ARRAY_OVERHEAD)


# propagate revert message when calls to external contracts fail
def check_external_call(call_ir):
    copy_revertdata = ["returndatacopy", 0, 0, "returndatasize"]
    revert = IRnode.from_list(["revert", 0, "returndatasize"], error_msg="external call failed")

    propagate_revert_ir = ["seq", copy_revertdata, revert]
    return ["if", ["iszero", call_ir], propagate_revert_ir]


# cost per byte of the identity precompile
def _identity_gas_bound(num_bytes):
    return GAS_IDENTITY + GAS_IDENTITYWORD * (ceil32(num_bytes) // 32)


def _mcopy_gas_bound(num_bytes):
    return GAS_COPY_WORD * ceil32(num_bytes) // 32


def _calldatacopy_gas_bound(num_bytes):
    return GAS_COPY_WORD * ceil32(num_bytes) // 32


def _codecopy_gas_bound(num_bytes):
    return GAS_COPY_WORD * ceil32(num_bytes) // 32


def data_location_to_address_space(s: DataLocation, is_ctor_ctx: bool) -> AddrSpace:
    if s == DataLocation.MEMORY:
        return MEMORY
    if s == DataLocation.STORAGE:
        return STORAGE
    if s == DataLocation.TRANSIENT:
        return TRANSIENT
    if s == DataLocation.CODE:
        if is_ctor_ctx:
            return IMMUTABLES
        return DATA

    raise CompilerPanic("unreachable!")  # pragma: nocover


def address_space_to_data_location(s: AddrSpace) -> DataLocation:
    if s == MEMORY:
        return DataLocation.MEMORY
    if s == STORAGE:
        return DataLocation.STORAGE
    if s == TRANSIENT:
        return DataLocation.TRANSIENT
    if s in (IMMUTABLES, DATA):
        return DataLocation.CODE
    if s == CALLDATA:
        return DataLocation.CALLDATA

    raise CompilerPanic("unreachable!")  # pragma: nocover


# Copy byte array word-for-word (including layout)
# TODO make this a private function
def make_byte_array_copier(dst, src):
    assert isinstance(src.typ, _BytestringT)
    assert isinstance(dst.typ, _BytestringT)

    _check_assign_bytes(dst, src)

    # TODO: remove this branch, copy_bytes and get_bytearray_length should handle
    if src.value == "~empty" or src.typ.maxlen == 0:
        # set length word to 0.
        return STORE(dst, 0)

    with src.cache_when_complex("src") as (b1, src):
        has_storage = STORAGE in (src.location, dst.location)
        is_memory_copy = dst.location == src.location == MEMORY
        batch_uses_identity = is_memory_copy and not version_check(begin="cancun")
        if src.typ.maxlen <= 32 and (has_storage or batch_uses_identity):
            # it's cheaper to run two load/stores instead of copy_bytes

            ret = ["seq"]
            # store length word
            len_ = get_bytearray_length(src)
            ret.append(STORE(dst, len_))

            # store the single data word.
            dst_data_ptr = bytes_data_ptr(dst)
            src_data_ptr = bytes_data_ptr(src)
            ret.append(STORE(dst_data_ptr, LOAD(src_data_ptr)))
            return b1.resolve(ret)

        # batch copy the bytearray (including length word) using copy_bytes
        len_ = add_ofst(get_bytearray_length(src), 32)
        max_bytes = src.typ.maxlen + 32
        ret = copy_bytes(dst, src, len_, max_bytes)
        return b1.resolve(ret)


def bytes_data_ptr(ptr):
    if ptr.location is None:  # pragma: nocover
        raise CompilerPanic("tried to modify non-pointer type")
    assert isinstance(ptr.typ, _BytestringT)
    return add_ofst(ptr, ptr.location.word_scale)


def dynarray_data_ptr(ptr):
    if ptr.location is None:  # pragma: nocover
        raise CompilerPanic("tried to modify non-pointer type")
    assert isinstance(ptr.typ, DArrayT)
    return add_ofst(ptr, ptr.location.word_scale)


def _dynarray_make_setter(dst, src):
    assert isinstance(src.typ, DArrayT)
    assert isinstance(dst.typ, DArrayT)

    if src.value == "~empty":
        return IRnode.from_list(STORE(dst, 0))

    # copy contents of src dynarray to dst.
    # note that in case src and dst refer to the same dynarray,
    # in order for get_element_ptr oob checks on the src dynarray
    # to work, we need to wait until after the data is copied
    # before we clobber the length word.

    if src.value == "multi":
        ret = ["seq"]
        # handle literals

        # copy each item
        n_items = len(src.args)

        for i in range(n_items):
            k = IRnode.from_list(i, typ=UINT256_T)
            dst_i = get_element_ptr(dst, k, array_bounds_check=False)
            src_i = get_element_ptr(src, k, array_bounds_check=False)
            ret.append(make_setter(dst_i, src_i))

        # write the length word after data is copied
        store_length = STORE(dst, n_items)
        ann = None
        if src.annotation is not None:
            ann = f"len({src.annotation})"
        store_length = IRnode.from_list(store_length, annotation=ann)

        ret.append(store_length)

        return ret

    with src.cache_when_complex("darray_src") as (b1, src):
        # for ABI-encoded dynamic data, we must loop to unpack, since
        # the layout does not match our memory layout
        should_loop = src.encoding == Encoding.ABI and src.typ.value_type.abi_type.is_dynamic()

        # if the data is not validated, we must loop to unpack
        should_loop |= needs_clamp(src.typ.value_type, src.encoding)

        # performance: if the subtype is dynamic, there might be a lot
        # of unused space inside of each element. for instance
        # DynArray[DynArray[uint256, 100], 5] where all the child
        # arrays are empty - for this case, we recursively call
        # into make_setter instead of straight bytes copy
        # TODO we can make this heuristic more precise, e.g.
        # loop when subtype.is_dynamic AND location == storage
        # OR array_size <= /bound where loop is cheaper than memcpy/
        should_loop |= src.typ.value_type.abi_type.is_dynamic()

        with get_dyn_array_count(src).cache_when_complex("darray_count") as (b2, count):
            ret = ["seq"]

            if should_loop:
                i = IRnode.from_list(_freshname("copy_darray_ix"), typ=UINT256_T)

                loop_body = make_setter(
                    get_element_ptr(dst, i, array_bounds_check=False),
                    get_element_ptr(src, i, array_bounds_check=False),
                )
                loop_body.annotation = f"{dst}[i] = {src}[i]"

                ret.append(["repeat", i, 0, count, src.typ.count, loop_body])
                # write the length word after data is copied
                ret.append(STORE(dst, count))

            else:
                element_size = src.typ.value_type.memory_bytes_required
                # number of elements * size of element in bytes + length word
                n_bytes = add_ofst(_mul(count, element_size), 32)
                max_bytes = 32 + src.typ.count * element_size

                # batch copy the entire dynarray, including length word
                ret.append(copy_bytes(dst, src, n_bytes, max_bytes))

            return b1.resolve(b2.resolve(ret))


# Copy bytes
# Accepts 4 arguments:
# (i) an IR node for the start position of the source
# (ii) an IR node for the start position of the destination
# (iii) an IR node for the length (in bytes)
# (iv) a constant for the max length (in bytes)
# NOTE: may pad to ceil32 of `length`! If you ask to copy 1 byte, it may
# copy an entire (32-byte) word, depending on the copy routine chosen.
# TODO maybe always pad to ceil32, to reduce dirty bytes bugs
def copy_bytes(dst, src, length, length_bound):
    annotation = f"copy up to {length_bound} bytes from {src} to {dst}"

    src = IRnode.from_list(src)
    dst = IRnode.from_list(dst)
    length = IRnode.from_list(length)

    with src.cache_when_complex("src") as (b1, src), length.cache_when_complex(
        "copy_bytes_count"
    ) as (b2, length), dst.cache_when_complex("dst") as (b3, dst):
        assert isinstance(length_bound, int) and length_bound >= 0

        # correctness: do not clobber dst
        if length_bound == 0:
            return IRnode.from_list(["seq"], annotation=annotation)
        # performance: if we know that length is 0, do not copy anything
        if length.value == 0:
            return IRnode.from_list(["seq"], annotation=annotation)

        assert src.is_pointer and dst.is_pointer

        # fast code for common case where num bytes is small
        if length_bound <= 32:
            copy_op = STORE(dst, LOAD(src))
            ret = IRnode.from_list(copy_op, annotation=annotation)
            return b1.resolve(b2.resolve(b3.resolve(ret)))

        if dst.location == MEMORY and src.location in (MEMORY, CALLDATA, DATA):
            # special cases: batch copy to memory
            # TODO: iloadbytes
            if src.location == MEMORY:
                if version_check(begin="cancun"):
                    copy_op = ["mcopy", dst, src, length]
                    gas_bound = _mcopy_gas_bound(length_bound)
                else:
                    copy_op = ["staticcall", "gas", 4, src, length, dst, length]
                    gas_bound = _identity_gas_bound(length_bound)
            elif src.location == CALLDATA:
                copy_op = ["calldatacopy", dst, src, length]
                gas_bound = _calldatacopy_gas_bound(length_bound)
            elif src.location == DATA:
                copy_op = ["dloadbytes", dst, src, length]
                # note: dloadbytes compiles to CODECOPY
                gas_bound = _codecopy_gas_bound(length_bound)

            ret = IRnode.from_list(copy_op, annotation=annotation, add_gas_estimate=gas_bound)
            return b1.resolve(b2.resolve(b3.resolve(ret)))

        if dst.location == IMMUTABLES and src.location in (MEMORY, DATA):
            # TODO istorebytes-from-mem, istorebytes-from-calldata(?)
            # compile to identity, CODECOPY respectively.
            pass

        # general case, copy word-for-word
        # pseudocode for our approach (memory-storage as example):
        # for i in range(len, bound=MAX_LEN):
        #   sstore(_dst + i, mload(src + i * 32))
        i = IRnode.from_list(_freshname("copy_bytes_ix"), typ=UINT256_T)

        # optimized form of (div (ceil32 len) 32)
        n = ["div", ["add", 31, length], 32]
        n_bound = ceil32(length_bound) // 32

        dst_i = add_ofst(dst, _mul(i, dst.location.word_scale))
        src_i = add_ofst(src, _mul(i, src.location.word_scale))

        copy_one_word = STORE(dst_i, LOAD(src_i))

        main_loop = ["repeat", i, 0, n, n_bound, copy_one_word]

        return b1.resolve(
            b2.resolve(b3.resolve(IRnode.from_list(main_loop, annotation=annotation)))
        )


# get the number of bytes at runtime
def get_bytearray_length(arg):
    typ = UINT256_T

    # TODO: it would be nice to merge the implementations of get_bytearray_length and
    # get_dynarray_count
    if arg.value == "~empty":
        return IRnode.from_list(0, typ=typ)

    return IRnode.from_list(LOAD(arg), typ=typ)


# get the number of elements at runtime
def get_dyn_array_count(arg):
    assert isinstance(arg.typ, DArrayT)

    typ = UINT256_T

    if arg.value == "multi":
        return IRnode.from_list(len(arg.args), typ=typ)

    if arg.value == "~empty":
        # empty(DynArray[...])
        return IRnode.from_list(0, typ=typ)

    return IRnode.from_list(LOAD(arg), typ=typ)


def append_dyn_array(darray_node, elem_node):
    assert isinstance(darray_node.typ, DArrayT)

    assert darray_node.typ.count > 0, "jerk boy u r out"

    ret = ["seq"]
    with darray_node.cache_when_complex("darray") as (b1, darray_node):
        len_ = get_dyn_array_count(darray_node)
        with len_.cache_when_complex("old_darray_len") as (b2, len_):
            assertion = ["assert", ["lt", len_, darray_node.typ.count]]
            ret.append(IRnode.from_list(assertion, error_msg=f"{darray_node.typ} bounds check"))
            # NOTE: typechecks elem_node
            # NOTE skip array bounds check bc we already asserted len two lines up
            ret.append(
                make_setter(get_element_ptr(darray_node, len_, array_bounds_check=False), elem_node)
            )

            # store new length
            ret.append(STORE(darray_node, ["add", len_, 1]))
            return IRnode.from_list(b1.resolve(b2.resolve(ret)))


def pop_dyn_array(darray_node, return_popped_item):
    assert isinstance(darray_node.typ, DArrayT)
    assert darray_node.encoding == Encoding.VYPER
    ret = ["seq"]
    with darray_node.cache_when_complex("darray") as (b1, darray_node):
        old_len = clamp("gt", get_dyn_array_count(darray_node), 0)
        new_len = IRnode.from_list(["sub", old_len, 1], typ=UINT256_T)

        with new_len.cache_when_complex("new_len") as (b2, new_len):
            # store new length
            ret.append(STORE(darray_node, new_len))

            # NOTE skip array bounds check bc we already asserted len two lines up
            if return_popped_item:
                popped_item = get_element_ptr(darray_node, new_len, array_bounds_check=False)
                ret.append(popped_item)
                typ = popped_item.typ
                location = popped_item.location
            else:
                typ, location = None, None

            return IRnode.from_list(b1.resolve(b2.resolve(ret)), typ=typ, location=location)


# add an offset to a pointer, keeping location and encoding info
def add_ofst(ptr, ofst):
    ret = ["add", ptr, ofst]
    return IRnode.from_list(ret, location=ptr.location, encoding=ptr.encoding)


# shorthand util
def _mul(x, y):
    ret = ["mul", x, y]
    return IRnode.from_list(ret)


# Resolve pointer locations for ABI-encoded data
def _getelemptr_abi_helper(parent, member_t, ofst, clamp=True):
    member_abi_t = member_t.abi_type

    # ABI encoding has length word and then pretends length is not there
    # e.g. [[1,2]] is encoded as 0x01 <len> 0x20 <inner array ofst> <encode(inner array)>
    # note that inner array ofst is 0x20, not 0x40.
    if has_length_word(parent.typ):
        parent = add_ofst(parent, parent.location.word_scale * DYNAMIC_ARRAY_OVERHEAD)

    ofst_ir = add_ofst(parent, ofst)

    if member_abi_t.is_dynamic():
        # double dereference, according to ABI spec
        # TODO optimize special case: first dynamic item
        # offset is statically known.
        ofst_ir = add_ofst(parent, unwrap_location(ofst_ir))

    return IRnode.from_list(
        ofst_ir,
        typ=member_t,
        location=parent.location,
        encoding=parent.encoding,
        annotation=f"{parent}{ofst}",
    )


# TODO simplify this code, especially the ABI decoding
def _get_element_ptr_tuplelike(parent, key):
    typ = parent.typ
    assert is_tuple_like(typ)

    if isinstance(typ, StructT):
        assert isinstance(key, str)
        subtype = typ.member_types[key]
        attrs = list(typ.tuple_keys())
        index = attrs.index(key)
        annotation = key
    else:
        # TupleT
        assert isinstance(key, int)
        subtype = typ.member_types[key]
        attrs = list(typ.tuple_keys())
        index = key
        annotation = None

    # generated by empty() + make_setter
    if parent.value == "~empty":
        return IRnode.from_list("~empty", typ=subtype)

    if parent.value == "multi":
        assert parent.encoding != Encoding.ABI, "no abi-encoded literals"
        return parent.args[index]

    ofst = 0  # offset from parent start

    if parent.encoding == Encoding.ABI:
        if parent.location in (STORAGE, TRANSIENT):  # pragma: nocover
            raise CompilerPanic("storage variables should not be abi encoded")

        member_t = typ.member_types[attrs[index]]

        for i in range(index):
            member_abi_t = typ.member_types[attrs[i]].abi_type
            ofst += member_abi_t.embedded_static_size()

        return _getelemptr_abi_helper(parent, member_t, ofst)

    data_location = address_space_to_data_location(parent.location)
    for i in range(index):
        t = typ.member_types[attrs[i]]
        ofst += t.get_size_in(data_location)

    return IRnode.from_list(
        add_ofst(parent, ofst),
        typ=subtype,
        location=parent.location,
        encoding=parent.encoding,
        annotation=annotation,
    )


def has_length_word(typ):
    # Consider moving this to an attribute on typ
    return isinstance(typ, (DArrayT, _BytestringT))


# TODO simplify this code, especially the ABI decoding
def _get_element_ptr_array(parent, key, array_bounds_check):
    assert is_array_like(parent.typ)

    if not is_integer_type(key.typ):  # pragma: nocover
        raise TypeCheckFailure(f"{key.typ} used as array index")

    subtype = parent.typ.value_type

    if parent.value == "~empty":
        if array_bounds_check:
            # this case was previously missing a bounds check. codegen
            # is a bit complicated when bounds check is required, so
            # block it. there is no reason to index into a literal empty
            # array anyways!
            raise TypeCheckFailure("indexing into zero array not allowed")
        return IRnode.from_list("~empty", subtype)

    if parent.value == "multi":
        assert isinstance(key.value, int)
        return parent.args[key.value]

    ix = unwrap_location(key)

    if array_bounds_check:
        is_darray = isinstance(parent.typ, DArrayT)
        bound = get_dyn_array_count(parent) if is_darray else parent.typ.count
        # NOTE: there are optimization rules for the bounds check when
        # ix or bound is literal
        with ix.cache_when_complex("ix") as (b1, ix):
            LT = "slt" if ix.typ.is_signed else "lt"
            # note: this is optimized out for unsigned integers
            is_negative = [LT, ix, 0]
            # always use unsigned ge, since bound is always an unsigned quantity
            is_oob = ["ge", ix, bound]
            checked_ix = ["seq", ["assert", ["iszero", ["or", is_negative, is_oob]]], ix]
            ix = b1.resolve(IRnode.from_list(checked_ix))
        ix.set_error_msg(f"{parent.typ} bounds check")

    if parent.encoding == Encoding.ABI:
        if parent.location in (STORAGE, TRANSIENT):  # pragma: nocover
            raise CompilerPanic("storage variables should not be abi encoded")

        member_abi_t = subtype.abi_type

        ofst = _mul(ix, member_abi_t.embedded_static_size())

        return _getelemptr_abi_helper(parent, subtype, ofst)

    data_location = address_space_to_data_location(parent.location)
    element_size = subtype.get_size_in(data_location)

    ofst = _mul(ix, element_size)

    if has_length_word(parent.typ):
        data_ptr = add_ofst(parent, parent.location.word_scale * DYNAMIC_ARRAY_OVERHEAD)
    else:
        data_ptr = parent

    return IRnode.from_list(add_ofst(data_ptr, ofst), typ=subtype, location=parent.location)


def _get_element_ptr_mapping(parent, key):
    assert isinstance(parent.typ, HashMapT)
    subtype = parent.typ.value_type
    key = unwrap_location(key)

    if parent.location not in (STORAGE, TRANSIENT):  # pragma: nocover
        raise TypeCheckFailure(f"bad dereference on mapping {parent}[{key}]")

    return IRnode.from_list(["sha3_64", parent, key], typ=subtype, location=parent.location)


# Take a value representing a memory or storage location, and descend down to
# an element or member variable
# This is analogous (but not necessarily equivalent to) getelementptr in LLVM.
def get_element_ptr(parent, key, array_bounds_check=True):
    with parent.cache_when_complex("val") as (b, parent):
        typ = parent.typ

        if is_tuple_like(typ):
            ret = _get_element_ptr_tuplelike(parent, key)

        elif isinstance(typ, HashMapT):
            ret = _get_element_ptr_mapping(parent, key)

        elif is_array_like(typ):
            ret = _get_element_ptr_array(parent, key, array_bounds_check)

        else:  # pragma: nocover
            raise CompilerPanic(f"get_element_ptr cannot be called on {typ}")

        return b.resolve(ret)


def LOAD(ptr: IRnode) -> IRnode:
    if ptr.location is None:  # pragma: nocover
        raise CompilerPanic("cannot dereference non-pointer type")
    op = ptr.location.load_op
    if op is None:  # pragma: nocover
        raise CompilerPanic(f"unreachable {ptr.location}")
    return IRnode.from_list([op, ptr])


def eval_once_check(name):
    # an IRnode which enforces uniqueness. include with a side-effecting
    # operation to sanity check that the codegen pipeline only generates
    # the side-effecting operation once (otherwise, IR-to-assembly will
    # throw a duplicate label exception). there is no runtime overhead
    # since the jumpdest gets optimized out in the final stage of assembly.
    return IRnode.from_list(["unique_symbol", name])


def ensure_eval_once(name, irnode):
    return ["seq", eval_once_check(_freshname(name)), irnode]


def STORE(ptr: IRnode, val: IRnode) -> IRnode:
    if ptr.location is None:  # pragma: nocover
        raise CompilerPanic("cannot dereference non-pointer type")
    op = ptr.location.store_op
    if op is None:  # pragma: nocover
        raise CompilerPanic(f"unreachable {ptr.location}")

    store = [op, ptr, val]
    # don't use eval_once_check for memory, immutables because it interferes
    # with optimizer
    if ptr.location in (MEMORY, IMMUTABLES):
        return IRnode.from_list(store)

    return IRnode.from_list(ensure_eval_once(f"{op}_", store))


# Unwrap location
def unwrap_location(orig):
    if orig.location is not None:
        return IRnode.from_list(LOAD(orig), typ=orig.typ)
    else:
        # CMC 2022-03-24 TODO refactor so this branch can be removed
        if orig.value == "~empty":
            # must be word type
            return IRnode.from_list(0, typ=orig.typ)
        return orig


# utility function, constructs an IR tuple out of a list of IR nodes
def ir_tuple_from_args(args):
    typ = TupleT([x.typ for x in args])
    return IRnode.from_list(["multi"] + [x for x in args], typ=typ)


def needs_external_call_wrap(typ):
    # for calls to ABI conforming contracts.
    # according to the ABI spec, return types are ALWAYS tuples even
    # if only one element is being returned.
    # https://solidity.readthedocs.io/en/latest/abi-spec.html#function-selector-and-argument-encoding
    # "and the return values v_1, ..., v_k of f are encoded as
    #
    #    enc((v_1, ..., v_k))
    #    i.e. the values are combined into a tuple and encoded.
    # "
    # therefore, wrap it in a tuple if it's not already a tuple.
    # for example, `bytes` is returned as abi-encoded (bytes,)
    # and `(bytes,)` is returned as abi-encoded ((bytes,),)
    # In general `-> X` gets returned as (X,)
    # including structs. MyStruct is returned as abi-encoded (MyStruct,).
    # (Sorry this is so confusing. I didn't make these rules.)

    return not (isinstance(typ, TupleT) and typ.length > 1)


def calculate_type_for_external_return(typ):
    if needs_external_call_wrap(typ):
        return TupleT([typ])
    return typ


def wrap_value_for_external_return(ir_val):
    # used for LHS promotion
    if needs_external_call_wrap(ir_val.typ):
        return ir_tuple_from_args([ir_val])
    else:
        return ir_val


def set_type_for_external_return(ir_val):
    # used for RHS promotion
    ir_val.typ = calculate_type_for_external_return(ir_val.typ)


# return a dummy IRnode with the given type
def dummy_node_for_type(typ):
    return IRnode("fake_node", typ=typ)


def _check_assign_bytes(left, right):
    if right.typ.maxlen > left.typ.maxlen:  # pragma: nocover
        raise TypeMismatch(f"Cannot cast from {right.typ} to {left.typ}")

    # stricter check for zeroing a byte array.
    # TODO: these should be TypeCheckFailure instead of TypeMismatch
    if right.value == "~empty" and right.typ.maxlen != left.typ.maxlen:  # pragma: nocover
        raise TypeMismatch(f"Cannot cast from empty({right.typ}) to {left.typ}")


def _check_assign_list(left, right):
    def FAIL():  # pragma: no cover
        raise TypeCheckFailure(f"assigning {right.typ} to {left.typ}")

    if left.value == "multi":  # pragma: nocover
        # Cannot do something like [a, b, c] = [1, 2, 3]
        FAIL()

    if isinstance(left.typ, SArrayT):
        if not is_array_like(right.typ):  # pragma: nocover
            FAIL()
        if left.typ.count != right.typ.count:  # pragma: nocover
            FAIL()

        # TODO recurse into left, right if literals?
        check_assign(
            dummy_node_for_type(left.typ.value_type), dummy_node_for_type(right.typ.value_type)
        )

    if isinstance(left.typ, DArrayT):
        if not isinstance(right.typ, DArrayT):  # pragma: nocover
            FAIL()

        if left.typ.count < right.typ.count:  # pragma: nocover
            FAIL()

        # stricter check for zeroing
        if right.value == "~empty" and right.typ.count != left.typ.count:  # pragma: nocover
            raise TypeCheckFailure(
                f"Bad type for clearing bytes: expected {left.typ} but got {right.typ}"
            )

        # TODO recurse into left, right if literals?
        check_assign(
            dummy_node_for_type(left.typ.value_type), dummy_node_for_type(right.typ.value_type)
        )


def _check_assign_tuple(left, right):
    def FAIL():  # pragma: no cover
        raise TypeCheckFailure(f"assigning {right.typ} to {left.typ}")

    if not isinstance(right.typ, left.typ.__class__):  # pragma: nocover
        FAIL()

    if isinstance(left.typ, StructT):
        for k in left.typ.member_types:
            if k not in right.typ.member_types:  # pragma: nocover
                FAIL()
            # TODO recurse into left, right if literals?
            check_assign(
                dummy_node_for_type(left.typ.member_types[k]),
                dummy_node_for_type(right.typ.member_types[k]),
            )

        for k in right.typ.member_types:
            if k not in left.typ.member_types:  # pragma: nocover
                FAIL()

        if left.typ.name != right.typ.name:  # pragma: nocover
            FAIL()

    else:
        if len(left.typ.member_types) != len(right.typ.member_types):  # pragma: nocover
            FAIL()
        for left_, right_ in zip(left.typ.member_types, right.typ.member_types):
            # TODO recurse into left, right if literals?
            check_assign(dummy_node_for_type(left_), dummy_node_for_type(right_))


# sanity check an assignment
# typechecking source code is done at an earlier phase
# this function is more of a sanity check for typechecking internally
# generated assignments
# TODO: do we still need this?
def check_assign(left, right):
    def FAIL():  # pragma: no cover
        raise TypeCheckFailure(f"assigning {right.typ} to {left.typ} {left} {right}")

    if isinstance(left.typ, _BytestringT):
        _check_assign_bytes(left, right)
    elif is_array_like(left.typ):
        _check_assign_list(left, right)
    elif is_tuple_like(left.typ):
        _check_assign_tuple(left, right)

    elif left.typ._is_prim_word:
        # TODO once we propagate types from typechecker, introduce this check:
        # if left.typ != right.typ:  # pragma: nocover
        #    FAIL()
        pass

    else:  # pragma: no cover
        FAIL()


_label = 0


# TODO might want to coalesce with Context.fresh_varname and compile_ir.mksymbol
def _freshname(name):
    global _label
    _label += 1
    return f"{name}{_label}"


def reset_names():
    global _label
    _label = 0


# returns True if t is ABI encoded and is a type that needs any kind of
# validation
def needs_clamp(t, encoding):
    if encoding == Encoding.VYPER:
        return False
    if encoding != Encoding.ABI:  # pragma: nocover
        raise CompilerPanic("unreachable")
    if isinstance(t, (_BytestringT, DArrayT)):
        return True
    if isinstance(t, FlagT):
        return len(t._flag_members) < 256
    if isinstance(t, SArrayT):
        return needs_clamp(t.value_type, encoding)
    if is_tuple_like(t):
        return any(needs_clamp(m, encoding) for m in t.tuple_members())
    if t._is_prim_word:
        return t not in (INT256_T, UINT256_T, BYTES32_T)

    raise CompilerPanic("unreachable")  # pragma: nocover


# Create an x=y statement, where the types may be compound
def make_setter(left, right):
    check_assign(left, right)

    # For types which occupy just one word we can use single load/store
    if left.typ._is_prim_word:
        enc = right.encoding  # unwrap_location butchers encoding
        right = unwrap_location(right)
        # TODO rethink/streamline the clamp_basetype logic
        if needs_clamp(right.typ, enc):
            right = clamp_basetype(right)

        return STORE(left, right)

    # Byte arrays
    elif isinstance(left.typ, _BytestringT):
        # TODO rethink/streamline the clamp_basetype logic
        if needs_clamp(right.typ, right.encoding):
            with right.cache_when_complex("bs_ptr") as (b, right):
                copier = make_byte_array_copier(left, right)
                ret = b.resolve(["seq", clamp_bytestring(right), copier])
        else:
            ret = make_byte_array_copier(left, right)

        return IRnode.from_list(ret)

    elif isinstance(left.typ, DArrayT):
        # TODO should we enable this?
        # implicit conversion from sarray to darray
        # if isinstance(right.typ, SArrayType):
        #    return _complex_make_setter(left, right)

        # TODO rethink/streamline the clamp_basetype logic
        if needs_clamp(right.typ, right.encoding):
            with right.cache_when_complex("arr_ptr") as (b, right):
                copier = _dynarray_make_setter(left, right)
                ret = b.resolve(["seq", clamp_dyn_array(right), copier])
        else:
            ret = _dynarray_make_setter(left, right)

        return IRnode.from_list(ret)

    # Complex Types
    assert isinstance(left.typ, (SArrayT, TupleT, StructT))

    return _complex_make_setter(left, right)


def _complex_make_setter(left, right):
    if right.value == "~empty" and left.location == MEMORY:
        # optimized memzero
        return mzero(left, left.typ.memory_bytes_required)

    ret = ["seq"]

    if isinstance(left.typ, SArrayT):
        n_items = right.typ.count
        keys = [IRnode.from_list(i, typ=UINT256_T) for i in range(n_items)]

    else:
        assert is_tuple_like(left.typ)
        keys = left.typ.tuple_keys()

    if left.is_pointer and right.is_pointer and right.encoding == Encoding.VYPER:
        # both left and right are pointers, see if we want to batch copy
        # instead of unrolling the loop.
        assert left.encoding == Encoding.VYPER
        len_ = left.typ.memory_bytes_required

        has_storage = STORAGE in (left.location, right.location)
        if has_storage:
            if _opt_codesize():
                # assuming PUSH2, a single sstore(dst (sload src)) is 8 bytes,
                # sstore(add (dst ofst), (sload (add (src ofst)))) is 16 bytes,
                # whereas loop overhead is 16-17 bytes.
                base_cost = 3
                if left._optimized.is_literal:
                    # code size is smaller since add is performed at compile-time
                    base_cost += 1
                if right._optimized.is_literal:
                    base_cost += 1
                # the formula is a heuristic, but it works.
                # (CMC 2023-07-14 could get more detailed for PUSH1 vs
                # PUSH2 etc but not worried about that too much now,
                # it's probably better to add a proper unroll rule in the
                # optimizer.)
                should_batch_copy = len_ >= 32 * base_cost
            elif _opt_gas():
                # kind of arbitrary, but cut off when code used > ~160 bytes
                should_batch_copy = len_ >= 32 * 10
            else:
                assert _opt_none()
                # don't care, just generate the most readable version
                should_batch_copy = True
        else:
            # find a cutoff for memory copy where identity is cheaper
            # than unrolled mloads/mstores
            # if MCOPY is available, mcopy is *always* better (except in
            # the 1 word case, but that is already handled by copy_bytes).
            if right.location == MEMORY and _opt_gas() and not version_check(begin="cancun"):
                # cost for 0th word - (mstore dst (mload src))
                base_unroll_cost = 12
                nth_word_cost = base_unroll_cost
                if not left._optimized.is_literal:
                    # (mstore (add N dst) (mload src))
                    nth_word_cost += 6
                if not right._optimized.is_literal:
                    # (mstore dst (mload (add N src)))
                    nth_word_cost += 6

                identity_base_cost = 115  # staticcall 4 gas dst len src len

                n_words = ceil32(len_) // 32
                should_batch_copy = (
                    base_unroll_cost + (nth_word_cost * (n_words - 1)) >= identity_base_cost
                )

            # calldata to memory, code to memory, cancun, or codesize -
            # batch copy is always better.
            else:
                should_batch_copy = True

        if should_batch_copy:
            return copy_bytes(left, right, len_, len_)

    # general case, unroll
    with left.cache_when_complex("_L") as (b1, left), right.cache_when_complex("_R") as (b2, right):
        for k in keys:
            l_i = get_element_ptr(left, k, array_bounds_check=False)
            r_i = get_element_ptr(right, k, array_bounds_check=False)
            ret.append(make_setter(l_i, r_i))

        return b1.resolve(b2.resolve(IRnode.from_list(ret)))


def ensure_in_memory(ir_var, context):
    """Ensure a variable is in memory. This is useful for functions
    which expect to operate on memory variables.
    """
    if ir_var.location == MEMORY:
        return ir_var

    typ = ir_var.typ
    buf = IRnode.from_list(context.new_internal_variable(typ), typ=typ, location=MEMORY)
    do_copy = make_setter(buf, ir_var)

    return IRnode.from_list(["seq", do_copy, buf], typ=typ, location=MEMORY)


def eval_seq(ir_node):
    """Tries to find the "return" value of a `seq` statement, in order so
    that the value can be known without possibly evaluating side effects
    """
    if ir_node.value in ("seq", "with") and len(ir_node.args) > 0:
        return eval_seq(ir_node.args[-1])
    if isinstance(ir_node.value, int):
        return IRnode.from_list(ir_node)
    return None


def mzero(dst, nbytes):
    # calldatacopy from past-the-end gives zero bytes.
    # cf. YP H.2 (ops section) with CALLDATACOPY spec.
    return IRnode.from_list(
        # calldatacopy mempos calldatapos len
        ["calldatacopy", dst, "calldatasize", nbytes],
        annotation="mzero",
    )


# zero pad a bytearray according to the ABI spec. The last word
# of the byte array needs to be right-padded with zeroes.
def zero_pad(bytez_placeholder):
    len_ = ["mload", bytez_placeholder]
    dst = ["add", ["add", bytez_placeholder, 32], "len"]
    # the runtime length of the data rounded up to nearest 32
    # from spec:
    #   the actual value of X as a byte sequence,
    #   followed by the *minimum* number of zero-bytes
    #   such that len(enc(X)) is a multiple of 32.
    # optimized form of ceil32(len) - len:
    num_zero_bytes = ["mod", ["sub", 0, "len"], 32]
    return IRnode.from_list(
        ["with", "len", len_, ["with", "dst", dst, mzero("dst", num_zero_bytes)]],
        annotation="Zero pad",
    )


# convenience rewrites for shr/sar/shl
def shr(bits, x):
    return ["shr", bits, x]


# convenience rewrites for shr/sar/shl
def shl(bits, x):
    return ["shl", bits, x]


def sar(bits, x):
    return ["sar", bits, x]


def clamp_bytestring(ir_node):
    t = ir_node.typ
    if not isinstance(t, _BytestringT):  # pragma: nocover
        raise CompilerPanic(f"{t} passed to clamp_bytestring")
    ret = ["assert", ["le", get_bytearray_length(ir_node), t.maxlen]]
    return IRnode.from_list(ret, error_msg=f"{ir_node.typ} bounds check")


def clamp_dyn_array(ir_node):
    t = ir_node.typ
    assert isinstance(t, DArrayT)
    ret = ["assert", ["le", get_dyn_array_count(ir_node), t.count]]
    return IRnode.from_list(ret, error_msg=f"{ir_node.typ} bounds check")


# clampers for basetype
def clamp_basetype(ir_node):
    t = ir_node.typ
    if not t._is_prim_word:  # pragma: nocover
        raise CompilerPanic(f"{t} passed to clamp_basetype")

    # copy of the input
    ir_node = unwrap_location(ir_node)

    if isinstance(t, FlagT):
        bits = len(t._flag_members)
        # assert x >> bits == 0
        ret = int_clamp(ir_node, bits, signed=False)

    elif isinstance(t, (IntegerT, DecimalT)):
        if t.bits == 256:
            ret = ir_node
        else:
            ret = int_clamp(ir_node, t.bits, signed=t.is_signed)

    elif isinstance(t, BytesM_T):
        if t.m == 32:
            ret = ir_node  # special case, no clamp.
        else:
            ret = bytes_clamp(ir_node, t.m)

    elif isinstance(t, (AddressT, InterfaceT)):
        ret = int_clamp(ir_node, 160)
    elif t in (BoolT(),):
        ret = int_clamp(ir_node, 1)
    else:  # pragma: no cover
        raise CompilerPanic(f"{t} passed to clamp_basetype")

    return IRnode.from_list(ret, typ=ir_node.typ, error_msg=f"validate {t}")


def int_clamp(ir_node, bits, signed=False):
    """Generalized clamper for integer types. Takes the number of bits,
    whether it's signed, and returns an IR node which checks it is
    in bounds. (Consumers should use clamp_basetype instead which uses
    type-based dispatch and is a little safer.)
    """
    if bits >= 256:  # pragma: nocover
        raise CompilerPanic(f"invalid clamp: {bits}>=256 ({ir_node})")

    u = "u" if not signed else ""
    msg = f"{u}int{bits} bounds check"
    with ir_node.cache_when_complex("val") as (b, val):
        if signed:
            # example for bits==128:
            # promote_signed_int(val, bits) is the "canonical" version of val
            # if val is in bounds, the bits above bit 128 should be equal.
            # (this works for both val >= 0 and val < 0. in the first case,
            # all upper bits should be 0 if val is a valid int128,
            # in the latter case, all upper bits should be 1.)
            assertion = ["assert", ["eq", val, promote_signed_int(val, bits)]]
        else:
            assertion = ["assert", ["iszero", shr(bits, val)]]

        assertion = IRnode.from_list(assertion, error_msg=msg)

        ret = b.resolve(["seq", assertion, val])

    return IRnode.from_list(ret, annotation=msg)


def bytes_clamp(ir_node: IRnode, n_bytes: int) -> IRnode:
    if not (0 < n_bytes <= 32):  # pragma: nocover
        raise CompilerPanic(f"bad type: bytes{n_bytes}")
    msg = f"bytes{n_bytes} bounds check"
    with ir_node.cache_when_complex("val") as (b, val):
        assertion = IRnode.from_list(["assert", ["iszero", shl(n_bytes * 8, val)]], error_msg=msg)
        ret = b.resolve(["seq", assertion, val])

    return IRnode.from_list(ret, annotation=msg)


# e.g. for int8, promote 255 to -1
def promote_signed_int(x, bits):
    assert bits % 8 == 0
    ret = ["signextend", bits // 8 - 1, x]
    return IRnode.from_list(ret, annotation=f"promote int{bits}")


# general clamp function for all ops and numbers
def clamp(op, arg, bound):
    with IRnode.from_list(arg).cache_when_complex("clamp_arg") as (b1, arg):
        check = IRnode.from_list(["assert", [op, arg, bound]], error_msg=f"clamp {op} {bound}")
        ret = ["seq", check, arg]
        return IRnode.from_list(b1.resolve(ret), typ=arg.typ)


def clamp_nonzero(arg):
    # TODO: use clamp("ne", arg, 0) once optimizer rules can handle it
    with IRnode.from_list(arg).cache_when_complex("should_nonzero") as (b1, arg):
        check = IRnode.from_list(["assert", arg], error_msg="check nonzero")
        ret = ["seq", check, arg]
        return IRnode.from_list(b1.resolve(ret), typ=arg.typ)


def clamp_le(arg, hi, signed):
    LE = "sle" if signed else "le"
    return clamp(LE, arg, hi)


def clamp2(lo, arg, hi, signed):
    with IRnode.from_list(arg).cache_when_complex("clamp2_arg") as (b1, arg):
        GE = "sge" if signed else "ge"
        LE = "sle" if signed else "le"
        ret = ["seq", ["assert", ["and", [GE, arg, lo], [LE, arg, hi]]], arg]
        return IRnode.from_list(b1.resolve(ret), typ=arg.typ)



import pytest

simple_cases = [
    (
        """
@external
def foo(t: bool, x: uint256, y: uint256) -> uint256:
    return x if t else y
    """,
        (1, 2),
    ),
    (  # literal test
        """
@external
def foo(_t: bool, x: uint256, y: uint256) -> uint256:
    return x if {test} else y
    """,
        (1, 2),
    ),
    (  # literal body
        """
@external
def foo(t: bool, _x: uint256, y: uint256) -> uint256:
    return {x} if t else y
    """,
        (1, 2),
    ),
    (  # literal orelse
        """
@external
def foo(t: bool, x: uint256, _y: uint256) -> uint256:
    return x if t else {y}
    """,
        (1, 2),
    ),
    (  # literal body/orelse
        """
@external
def foo(t: bool, _x: uint256, _y: uint256) -> uint256:
    return {x} if t else {y}
    """,
        (1, 2),
    ),
    (  # literal everything
        """
@external
def foo(_t: bool, _x: uint256, _y: uint256) -> uint256:
    return {x} if {test} else {y}
    """,
        (1, 2),
    ),
    (  # body/orelse in storage and memory
        """
s: uint256
@external
def foo(t: bool, x: uint256, y: uint256) -> uint256:
    self.s = x
    return self.s if t else y
    """,
        (1, 2),
    ),
    (  # body/orelse in memory and storage
        """
s: uint256
@external
def foo(t: bool, x: uint256, y: uint256) -> uint256:
    self.s = x
    return self.s if t else y
    """,
        (1, 2),
    ),
    (  # body/orelse in memory and constant
        """
S: constant(uint256) = {y}
@external
def foo(t: bool, x: uint256, _y: uint256) -> uint256:
    return x if t else S
    """,
        (1, 2),
    ),
    (  # dynarray
        """
@external
def foo(t: bool, x: DynArray[uint256, 3], y: DynArray[uint256, 3]) -> DynArray[uint256, 3]:
    return x if t else y
    """,
        ([], [1]),
    ),
    (  # variable + literal dynarray
        """
@external
def foo(t: bool, x: DynArray[uint256, 3], _y: DynArray[uint256, 3]) -> DynArray[uint256, 3]:
    return x if t else {y}
    """,
        ([], [1]),
    ),
    (  # literal + variable dynarray
        """
@external
def foo(t: bool, _x: DynArray[uint256, 3], y: DynArray[uint256, 3]) -> DynArray[uint256, 3]:
    return {x} if t else y
    """,
        ([], [1]),
    ),
    (  # storage dynarray
        """
s: DynArray[uint256, 3]
@external
def foo(t: bool, x: DynArray[uint256, 3], y: DynArray[uint256, 3]) -> DynArray[uint256, 3]:
    self.s = y
    return x if t else self.s
    """,
        ([], [1]),
    ),
    (  # static array
        """
@external
def foo(t: bool, x: uint256[1], y: uint256[1]) -> uint256[1]:
    return x if t else y
    """,
        ([2], [1]),
    ),
    (  # static array literal
        """
@external
def foo(t: bool, x: uint256[1], _y: uint256[1]) -> uint256[1]:
    return x if t else {y}
    """,
        ([2], [1]),
    ),
    (  # strings
        """
@external
def foo(t: bool, x: String[10], y: String[10]) -> String[10]:
    return x if t else y
    """,
        ("hello", "world"),
    ),
    (  # string literal
        """
@external
def foo(t: bool, x: String[10], _y: String[10]) -> String[10]:
    return x if t else {y}
    """,
        ("hello", "world"),
    ),
    (  # bytes
        """
@external
def foo(t: bool, x: Bytes[10], y: Bytes[10]) -> Bytes[10]:
    return x if t else y
    """,
        (b"hello", b"world"),
    ),
]


@pytest.mark.parametrize("code,inputs", simple_cases)
@pytest.mark.parametrize("test", [True, False])
def test_ternary_simple(get_contract, code, test, inputs):
    x, y = inputs
    # note: repr to escape strings
    code = code.format(test=test, x=repr(x), y=repr(y))
    c = get_contract(code)
    # careful with order of precedence of `assert` and `if/else` in python!
    assert c.foo(test, x, y) == (x if test else y)


tuple_codes = [
    """
@external
def foo(t: bool, x: uint256, y: uint256) -> (uint256, uint256):
    return (x, y) if t else (y, x)
    """,
    """
s: uint256
@external
def foo(t: bool, x: uint256, y: uint256) -> (uint256, uint256):
    self.s = x
    return (self.s, y) if t else (y, self.s)
    """,
]


@pytest.mark.parametrize("code", tuple_codes)
@pytest.mark.parametrize("test", [True, False])
def test_ternary_tuple(get_contract, code, test):
    c = get_contract(code)

    x, y = 1, 2
    assert c.foo(test, x, y) == ((x, y) if test else (y, x))


@pytest.mark.parametrize("test", [True, False])
def test_ternary_immutable(get_contract, test):
    code = """
IMM: public(immutable(uint256))
@deploy
def __init__(test: bool):
    IMM = 1 if test else 2
    """
    c = get_contract(code, test)

    assert c.IMM() == (1 if test else 2)


@pytest.mark.parametrize("test", [True, False])
@pytest.mark.parametrize("x", list(range(8)))
@pytest.mark.parametrize("y", list(range(8)))
def test_complex_ternary_expression(get_contract, test, x, y):
    code = """
@external
def foo(t: bool, x: uint256, y: uint256) -> uint256:
    return (x * y) if (t and True) else (x + y + convert(t, uint256))
    """
    c = get_contract(code)

    assert c.foo(test, x, y) == ((x * y) if (test and True) else (x + y + int(test)))


@pytest.mark.parametrize("test", [True, False])
@pytest.mark.parametrize("x", list(range(8)))
@pytest.mark.parametrize("y", list(range(8)))
def test_ternary_precedence(get_contract, test, x, y):
    code = """
@external
def foo(t: bool, x: uint256, y: uint256) -> uint256:
    return x * y if t else x + y + convert(t, uint256)
    """
    c = get_contract(code)

    assert c.foo(test, x, y) == (x * y if test else x + y + int(test))


@pytest.mark.parametrize("test1", [True, False])
@pytest.mark.parametrize("test2", [True, False])
def test_nested_ternary(get_contract, test1, test2):
    code = """
@external
def foo(t1: bool, t2: bool, x: uint256, y: uint256, z: uint256) -> uint256:
    return x if t1 else y if t2 else z
    """
    c = get_contract(code)

    x, y, z = 1, 2, 3
    assert c.foo(test1, test2, x, y, z) == (x if test1 else y if test2 else z)


@pytest.mark.parametrize("test", [True, False])
def test_ternary_side_effects(get_contract, test):
    code = """
track_taint_x: public(uint256)
track_taint_y: public(uint256)
foo_retval: public(uint256)

@internal
def x() -> uint256:
    self.track_taint_x += 1
    return 5

@internal
def y() -> uint256:
    self.track_taint_y += 1
    return 7

@external
def foo(t: bool):
    self.foo_retval = self.x() if t else self.y()
    """
    c = get_contract(code)

    c.foo(test)
    assert c.foo_retval() == (5 if test else 7)

    if test:
        assert c.track_taint_x() == 1
        assert c.track_taint_y() == 0
    else:
        assert c.track_taint_x() == 0
        assert c.track_taint_y() == 1



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import NamespaceCollision, UndeclaredDefinition

fail_list = [
    """
@external
def foo(choice: bool):
    if (choice):
        a: int128 = 1
    a += 1
    """,
    """
@external
def foo(choice: bool):
    if (choice):
        a: int128 = 0
    else:
        a: int128 = 1
    a += 1
    """,
    """
@external
def foo(choice: bool):
    if (choice):
        a: int128 = 0
    else:
        a += 1
    """,
    """
@external
def foo(choice: bool):

    for i: int128 in range(4):
        a: int128 = 0
    a = 1
    """,
    """
@external
def foo(choice: bool):

    for i: int128 in range(4):
        a: int128 = 0
    a += 1
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_fail_undeclared(bad_code):
    with raises(UndeclaredDefinition):
        compiler.compile_code(bad_code)


fail_list_collision = [
    """
@external
def foo():
    a: int128 = 5
    a: int128 = 7
    """
]


@pytest.mark.parametrize("bad_code", fail_list_collision)
def test_fail_collision(bad_code):
    with raises(NamespaceCollision):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(choice: bool, choice2: bool):
    if (choice):
        a: int128 = 11
        if choice2 and a > 1:
            a -= 1  # should be visible here.
    """,
    """
@external
def foo(choice: bool):
    if choice:
        a: int128 = 44
    else:
        a: uint256 = 42
    a: bool = True
    """,
    """
a: int128

@external
def foo():
    a: int128 = 5
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_valid_blockscope(good_code):
    assert compiler.compile_code(good_code) is not None



def test_repeat(get_contract_from_ir, assert_compile_failed):
    good_ir = ["repeat", 0, 0, 1, 1, ["seq"]]
    bad_ir_1 = ["repeat", 0, 0, 0, 0, ["seq"]]
    bad_ir_2 = ["repeat", 0, 0, -1, -1, ["seq"]]
    get_contract_from_ir(good_ir)
    assert_compile_failed(lambda: get_contract_from_ir(bad_ir_1), Exception)
    assert_compile_failed(lambda: get_contract_from_ir(bad_ir_2), Exception)



# EXPERIMENTAL VYPER PARSER
import textwrap

from lark import Lark
from lark.indenter import Indenter


class PythonIndenter(Indenter):
    NL_type = "_NEWLINE"
    OPEN_PAREN_types = ["LPAR", "LSQB", "LBRACE"]
    CLOSE_PAREN_types = ["RPAR", "RSQB", "RBRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 4


_lark_grammar = None


def vyper_grammar():
    global _lark_grammar
    if _lark_grammar is None:
        _lark_grammar = Lark.open_from_package(
            "vyper",
            "grammar.lark",
            ("ast/",),
            parser="lalr",
            start="module",
            postlex=PythonIndenter(),
        )
    return _lark_grammar


def parse_vyper_source(code, dedent=False):
    if dedent:
        code = textwrap.dedent(code)
    return vyper_grammar().parse(code + "\n")



import pytest

from vyper import compiler

valid_list = [
    """
@external
def foo(x: uint256):
    print(x)
    """,
    """
@external
def foo(x: Bytes[1]):
    print(x)
    """,
    """
struct Foo:
    x: Bytes[128]
@external
def foo(foo: Foo):
    print(foo)
    """,
    """
struct Foo:
    x: uint256
@external
def foo(foo: Foo):
    print(foo)
    """,
    """
BAR: constant(DynArray[uint256, 5]) = [1, 2, 3, 4, 5]

@external
def foo():
    print(BAR)
    """,
    """
FOO: constant(uint256) = 1
BAR: constant(DynArray[uint256, 5]) = [1, 2, 3, 4, 5]

@external
def foo():
    print(FOO, BAR)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_print_syntax(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo() -> int128:
    return -2**127
    """,
        TypeMismatch,
    )
]


@pytest.mark.parametrize("code,exc", fail_list)
def test_unary_fail(code, exc):
    with pytest.raises(exc):
        compile_code(code)



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import StructureException, SyntaxException, UnknownType


def test_external_contract_call_declaration_expr(get_contract, tx_failed):
    contract_1 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128) -> int128:
    self.lucky = _lucky
    return self.lucky
"""

    contract_2 = """
interface ModBar:
    def set_lucky(_lucky: int128) -> int128: nonpayable

interface ConstBar:
    def set_lucky(_lucky: int128) -> int128: view

modifiable_bar_contract: ModBar
static_bar_contract: ConstBar

@deploy
def __init__(contract_address: address):
    self.modifiable_bar_contract = ModBar(contract_address)
    self.static_bar_contract = ConstBar(contract_address)

@external
def modifiable_set_lucky(_lucky: int128):
    extcall self.modifiable_bar_contract.set_lucky(_lucky)

@external
def static_set_lucky(_lucky: int128):
    s: int128 = staticcall self.static_bar_contract.set_lucky(_lucky)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2, *[c1.address])
    c2.modifiable_set_lucky(7)
    assert c1.lucky() == 7
    # Fails attempting a state change after a call to a static address
    with tx_failed():
        c2.static_set_lucky(5)
    assert c1.lucky() == 7


def test_external_contract_call_declaration_stmt(get_contract, tx_failed):
    contract_1 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128) -> int128:
    self.lucky = _lucky
    return self.lucky
"""

    contract_2 = """
interface ModBar:
    def set_lucky(_lucky: int128) -> int128: nonpayable

interface ConstBar:
    def set_lucky(_lucky: int128) -> int128: view

modifiable_bar_contract: ModBar
static_bar_contract: ConstBar

@deploy
def __init__(contract_address: address):
    self.modifiable_bar_contract = ModBar(contract_address)
    self.static_bar_contract = ConstBar(contract_address)

@external
def modifiable_set_lucky(_lucky: int128) -> int128:
    x: int128 = extcall self.modifiable_bar_contract.set_lucky(_lucky)
    return x

@external
def static_set_lucky(_lucky: int128):
    x:int128 = staticcall self.static_bar_contract.set_lucky(_lucky)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2, *[c1.address])
    c2.modifiable_set_lucky(7)
    assert c1.lucky() == 7
    # Fails attempting a state change after a call to a static address
    with tx_failed():
        c2.static_set_lucky(5)
    assert c1.lucky() == 7


def test_multiple_contract_state_changes(get_contract, tx_failed):
    contract_1 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128) -> int128:
    self.lucky = _lucky
    return self.lucky
"""

    contract_2 = """
interface ModBar:
    def set_lucky(_lucky: int128) -> int128: nonpayable

interface ConstBar:
    def set_lucky(_lucky: int128) -> int128: view

modifiable_bar_contract: ModBar
static_bar_contract: ConstBar

@deploy
def __init__(contract_address: address):
    self.modifiable_bar_contract = ModBar(contract_address)
    self.static_bar_contract = ConstBar(contract_address)

@external
def modifiable_set_lucky(_lucky: int128) -> int128:
    return extcall self.modifiable_bar_contract.set_lucky(_lucky)

@external
def static_set_lucky(_lucky: int128) -> int128:
    return staticcall self.static_bar_contract.set_lucky(_lucky)
    """

    contract_3 = """
interface ModBar:
    def modifiable_set_lucky(_lucky: int128) -> int128: nonpayable
    def static_set_lucky(_lucky: int128) -> int128: nonpayable

interface ConstBar:
    def modifiable_set_lucky(_lucky: int128) -> int128: view
    def static_set_lucky(_lucky: int128) -> int128: view

modifiable_bar_contract: ModBar
static_bar_contract: ConstBar

@deploy
def __init__(contract_address: address):
    self.modifiable_bar_contract = ModBar(contract_address)
    self.static_bar_contract = ConstBar(contract_address)

@external
def modifiable_modifiable_set_lucky(_lucky: int128):
    extcall self.modifiable_bar_contract.modifiable_set_lucky(_lucky)

@external
def modifiable_static_set_lucky(_lucky: int128):
    extcall self.modifiable_bar_contract.static_set_lucky(_lucky)

@external
def static_static_set_lucky(_lucky: int128) -> int128:
    return staticcall self.static_bar_contract.static_set_lucky(_lucky)

@external
def static_modifiable_set_lucky(_lucky: int128) -> int128:
    return staticcall self.static_bar_contract.modifiable_set_lucky(_lucky)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2, *[c1.address])
    c3 = get_contract(contract_3, *[c2.address])

    assert c1.lucky() == 0
    c3.modifiable_modifiable_set_lucky(7)
    assert c1.lucky() == 7
    with tx_failed():
        c3.modifiable_static_set_lucky(6)
    with tx_failed():
        c3.static_modifiable_set_lucky(6)
    with tx_failed():
        c3.static_static_set_lucky(6)
    assert c1.lucky() == 7


def test_address_can_returned_from_contract_type(get_contract):
    contract_1 = """
@external
def bar() -> int128:
    return 1
"""
    contract_2 = """
interface Bar:
    def bar() -> int128: view

bar_contract: public(Bar)

@external
def foo(contract_address: address):
    self.bar_contract = Bar(contract_address)

@external
def get_bar() -> int128:
    return staticcall self.bar_contract.bar()
"""
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    c2.foo(c1.address)
    assert c2.bar_contract() == c1.address
    assert c2.get_bar() == 1


def test_invalid_external_contract_call_declaration_1():
    contract_1 = """
interface Bar:
    def bar() -> int128: pass
    """

    with pytest.raises(StructureException):
        compile_code(contract_1)


def test_invalid_external_contract_call_declaration_2():
    contract_1 = """
interface Bar:
    def bar() -> int128: view

bar_contract: Boo

@external
def foo(contract_address: address) -> int128:
    self.bar_contract = Bar(contract_address)
    return self.bar_contract.bar()
    """

    with pytest.raises(UnknownType):
        compile_code(contract_1)


def test_invalid_if_external_contract_doesnt_exist():
    code = """
modifiable_bar_contract: Bar
    """

    with pytest.raises(UnknownType):
        compile_code(code)


def test_invalid_if_not_in_valid_global_keywords():
    code = """
interface Bar:
    def set_lucky(_lucky: int128): nonpayable

modifiable_bar_contract: trusted(Bar)
    """
    with pytest.raises(SyntaxException):
        compile_code(code)


def test_invalid_if_have_modifiability_not_declared():
    code = """
interface Bar:
    def set_lucky(_lucky: int128): pass
    """
    with pytest.raises(StructureException):
        compile_code(code)



import pytest

from tests.utils import ZERO_ADDRESS
from vyper import compile_code
from vyper.exceptions import TypeMismatch


def test_correct_abi_right_padding(env, get_contract):
    selfcall_code_6 = """
@external
def hardtest(arg1: Bytes[64], arg2: Bytes[64]) -> Bytes[128]:
    return concat(arg1, arg2)
    """

    c = get_contract(selfcall_code_6)

    assert c.hardtest(b"hello" * 5, b"hello" * 10) == b"hello" * 15

    res = env.message_call(
        to=c.address, data=c.hardtest.prepare_calldata(b"hello" * 5, b"hello" * 10)
    )
    static_offset = int.from_bytes(res[:32], "big")
    assert static_offset == 32

    dyn_section = res[static_offset:]
    assert len(dyn_section) % 32 == 0  # first right pad assert

    len_value = int.from_bytes(dyn_section[:32], "big")

    assert len_value == len(b"hello" * 15)
    assert dyn_section[32 : 32 + len_value] == b"hello" * 15
    # second right pad assert
    assert dyn_section[32 + len_value :] == b"\x00" * (len(dyn_section) - 32 - len_value)


def test_return_type(get_contract):
    long_string = 35 * "test"

    code = """
struct Chunk:
    a: Bytes[8]
    b: Bytes[8]
    c: int128
chunk: Chunk

@deploy
def __init__():
    self.chunk.a = b"hello"
    self.chunk.b = b"world"
    self.chunk.c = 5678

@external
def out() -> (int128, address):
    return 3333, 0x0000000000000000000000000000000000000001

@external
def out_literals() -> (int128, address, Bytes[6]):
    return 1, 0x0000000000000000000000000000000000000000, b"random"

@external
def out_bytes_first() -> (Bytes[4], int128):
    return b"test", 1234

@external
def out_bytes_a(x: int128, y: Bytes[4]) -> (int128, Bytes[4]):
    return x, y

@external
def out_bytes_b(x: int128, y: Bytes[4]) -> (Bytes[4], int128, Bytes[4]):
    return y, x, y

@external
def four() -> (int128, Bytes[8], Bytes[8], int128):
    return 1234, b"bytes", b"test", 4321

@external
def out_chunk() -> (Bytes[8], int128, Bytes[8]):
    return self.chunk.a, self.chunk.c, self.chunk.b

@external
def out_very_long_bytes() -> (int128, Bytes[1024], int128, address):
    return 5555, b"testtesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttesttest", 6666, 0x0000000000000000000000000000000000001234  # noqa
    """

    c = get_contract(code)

    assert c.out() == (3333, "0x0000000000000000000000000000000000000001")
    assert c.out_literals() == (1, ZERO_ADDRESS, b"random")
    assert c.out_bytes_first() == (b"test", 1234)
    assert c.out_bytes_a(5555555, b"test") == (5555555, b"test")
    assert c.out_bytes_b(5555555, b"test") == (b"test", 5555555, b"test")
    assert c.four() == (1234, b"bytes", b"test", 4321)
    assert c.out_chunk() == (b"hello", 5678, b"world")
    assert c.out_very_long_bytes() == (
        5555,
        long_string.encode(),
        6666,
        "0x0000000000000000000000000000000000001234",
    )


def test_return_type_signatures(get_contract):
    code = """
@external
def out_literals() -> (int128, address, Bytes[6]):
    return 1, 0x0000000000000000000000000000000000000000, b"random"
    """

    c = get_contract(code)
    assert c.abi[0]["outputs"] == [
        {"type": "int128", "name": ""},
        {"type": "address", "name": ""},
        {"type": "bytes", "name": ""},
    ]


def test_return_tuple_assign(get_contract):
    code = """
@internal
def _out_literals() -> (int128, address, Bytes[10]):
    return 1, 0x0000000000000000000000000000000000000000, b"random"

@external
def out_literals() -> (int128, address, Bytes[10]):
    return self._out_literals()

@external
def test() -> (int128, address, Bytes[10]):
    a: int128 = 0
    b: address = empty(address)
    c: Bytes[10] = b""
    (a, b, c) = self._out_literals()
    return a, b, c
    """

    c = get_contract(code)

    assert c.out_literals() == c.test() == (1, ZERO_ADDRESS, b"random")


def test_return_tuple_assign_storage(get_contract):
    code = """
a: int128
b: address
c: Bytes[20]
d: Bytes[20]

@internal
def _out_literals() -> (int128, Bytes[20], address, Bytes[20]):
    return 1, b"testtesttest", 0x0000000000000000000000000000000000000023, b"random"

@external
def out_literals() -> (int128, Bytes[20], address, Bytes[20]):
    return self._out_literals()

@external
def test1() -> (int128, Bytes[20], address, Bytes[20]):
    self.a, self.c, self.b, self.d = self._out_literals()
    return self.a, self.c, self.b, self.d

@external
def test2() -> (int128, address):
    x: int128 = 0
    x, self.c, self.b, self.d = self._out_literals()
    return x, self.b

@external
def test3() -> (address, int128):
    x: address = empty(address)
    self.a, self.c, x, self.d = self._out_literals()
    return x, self.a
    """

    c = get_contract(code)

    addr = "0x" + "00" * 19 + "23"
    assert c.out_literals() == (1, b"testtesttest", addr, b"random")
    assert c.out_literals() == c.test1()
    assert c.test2() == (1, c.out_literals()[2])
    assert c.test3() == (c.out_literals()[2], 1)


@pytest.mark.parametrize("string", ["a", "abc", "abcde", "potato"])
def test_string_inside_tuple(get_contract, string):
    code = f"""
@external
def test_return() -> (String[6], uint256):
    return "{string}", 42
    """
    c1 = get_contract(code)

    code = """
interface jsonabi:
    def test_return() -> (String[6], uint256): view

@external
def test_values(a: address) -> (String[6], uint256):
    return staticcall jsonabi(a).test_return()
    """

    c2 = get_contract(code)
    assert c2.test_values(c1.address) == (string, 42)


@pytest.mark.parametrize("string", ["a", "abc", "abcde", "potato"])
def test_bytes_inside_tuple(get_contract, string):
    code = f"""
@external
def test_return() -> (Bytes[6], uint256):
    return b"{string}", 42
    """
    c1 = get_contract(code)

    code = """
interface jsonabi:
    def test_return() -> (Bytes[6], uint256): view

@external
def test_values(a: address) -> (Bytes[6], uint256):
    return staticcall jsonabi(a).test_return()
    """

    c2 = get_contract(code)
    assert c2.test_values(c1.address) == (bytes(string, "utf-8"), 42)


def test_tuple_return_typecheck(tx_failed, get_contract):
    code = """
@external
def getTimeAndBalance() -> (bool, address):
    return block.timestamp, self.balance
    """
    with pytest.raises(TypeMismatch):
        compile_code(code)


def test_struct_return_abi(get_contract):
    code = """
struct Voter:
    weight: int128
    voted: bool

@external
def test() -> Voter:
    a: Voter = Voter(weight=123, voted=True)
    return a
    """

    out = compile_code(code, output_formats=["abi"])
    abi = out["abi"][0]

    assert abi["name"] == "test"

    c = get_contract(code)

    assert c.test() == (123, True)


def test_single_struct_return_abi(get_contract):
    code = """
struct Voter:
    voted: bool

@external
def test() -> Voter:
    a: Voter = Voter(voted=True)
    return a
    """

    out = compile_code(code, output_formats=["abi"])
    abi = out["abi"][0]

    assert abi["name"] == "test"
    assert abi["outputs"][0]["type"] == "tuple"

    c = get_contract(code)

    assert c.test() == (True,)


def test_struct_return(get_contract):
    code = """
struct Foo:
  x: int128
  y: uint256

_foo: Foo
_foos: HashMap[int128, Foo]

@internal
def priv1() -> Foo:
    return Foo(x= 1, y=2)
@external
def pub1() -> Foo:
    return self.priv1()

@internal
def priv2() -> Foo:
    foo: Foo = Foo(x= 0, y=0)
    foo.x = 3
    foo.y = 4
    return foo
@external
def pub2() -> Foo:
    return self.priv2()

@external
def pub3() -> Foo:
    self._foo = Foo(x= 5, y=6)
    return self._foo

@external
def pub4() -> Foo:
   self._foos[0] = Foo(x= 7, y=8)
   return self._foos[0]

@internal
def return_arg(foo: Foo) -> Foo:
    return foo
@external
def pub5(foo: Foo) -> Foo:
    return self.return_arg(foo)
@external
def pub6() -> Foo:
    foo: Foo = Foo(x= 123, y=456)
    return self.return_arg(foo)
    """
    foo = (123, 456)

    c = get_contract(code)

    assert c.pub1() == (1, 2)
    assert c.pub2() == (3, 4)
    assert c.pub3() == (5, 6)
    assert c.pub4() == (7, 8)
    assert c.pub5(foo) == foo
    assert c.pub6() == foo


def test_single_struct_return(get_contract):
    code = """
struct Foo:
  x: int128

_foo: Foo
_foos: HashMap[int128, Foo]

@internal
def priv1() -> Foo:
    return Foo(x=1)
@external
def pub1() -> Foo:
    return self.priv1()

@internal
def priv2() -> Foo:
    foo: Foo = Foo(x=0)
    foo.x = 3
    return foo
@external
def pub2() -> Foo:
    return self.priv2()

@external
def pub3() -> Foo:
    self._foo = Foo(x=5)
    return self._foo

@external
def pub4() -> Foo:
   self._foos[0] = Foo(x=7)
   return self._foos[0]

@internal
def return_arg(foo: Foo) -> Foo:
    return foo
@external
def pub5(foo: Foo) -> Foo:
    return self.return_arg(foo)
@external
def pub6() -> Foo:
    foo: Foo = Foo(x=123)
    return self.return_arg(foo)
    """
    foo = (123,)

    c = get_contract(code)

    assert c.pub1() == (1,)
    assert c.pub2() == (3,)
    assert c.pub3() == (5,)
    assert c.pub4() == (7,)
    assert c.pub5(foo) == foo
    assert c.pub6() == foo


def test_self_call_in_return_struct(get_contract):
    code = """
struct Foo:
    a: uint256
    b: uint256
    c: uint256
    d: uint256
    e: uint256

@internal
def _foo() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 3

@external
def foo() -> Foo:
    return Foo(a=1, b=2, c=self._foo(), d=4, e=5)
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_self_call_in_return_single_struct(get_contract):
    code = """
struct Foo:
    a: uint256

@internal
def _foo() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 3

@external
def foo() -> Foo:
    return Foo(a=self._foo())
    """

    c = get_contract(code)

    assert c.foo() == (3,)


def test_call_in_call(get_contract):
    code = """
struct Foo:
    a: uint256
    b: uint256
    c: uint256
    d: uint256
    e: uint256

@internal
def _foo(a: uint256, b: uint256, c: uint256) -> Foo:
    return Foo(a=1, b=a, c=b, d=c, e=5)

@internal
def _foo2() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,15,16]
    return 4

@external
def foo() -> Foo:
    return self._foo(2, 3, self._foo2())
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_call_in_call_single_struct(get_contract):
    code = """
struct Foo:
    a: uint256

@internal
def _foo(a: uint256) -> Foo:
    return Foo(a=a)

@internal
def _foo2() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,15,16]
    return 4

@external
def foo() -> Foo:
    return self._foo(self._foo2())
    """

    c = get_contract(code)

    assert c.foo() == (4,)


def test_nested_calls_in_struct_return(get_contract):
    code = """
struct Foo:
    a: uint256
    b: uint256
    c: uint256
    d: uint256
    e: uint256
struct Bar:
    a: uint256
    b: uint256

@internal
def _bar(a: uint256, b: uint256, c: uint256) -> Bar:
    return Bar(a=415, b=3)

@internal
def _foo2(a: uint256) -> uint256:
    b: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 99

@internal
def _foo3(a: uint256, b: uint256) -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 42

@internal
def _foo4() -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 4

@external
def foo() -> Foo:
    return Foo(
        a=1,
        b=2,
        c=self._bar(6, 7, self._foo2(self._foo3(9, 11))).b,
        d=self._foo4(),
        e=5
    )
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_nested_calls_in_single_struct_return(get_contract):
    code = """
struct Foo:
    a: uint256
struct Bar:
    a: uint256
    b: uint256

@internal
def _bar(a: uint256, b: uint256, c: uint256) -> Bar:
    return Bar(a=415, b=3)

@internal
def _foo2(a: uint256) -> uint256:
    b: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 99

@internal
def _foo3(a: uint256, b: uint256) -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 42

@internal
def _foo4() -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 4

@external
def foo() -> Foo:
    return Foo(
        a=self._bar(6, self._foo4(), self._foo2(self._foo3(9, 11))).b,
    )
    """

    c = get_contract(code)

    assert c.foo() == (3,)


def test_external_call_in_return_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256
@view
@external
def bar() -> Bar:
    return Bar(a=3, b=4)
    """

    code2 = """
struct Foo:
    a: uint256
    b: uint256
    c: uint256
    d: uint256
    e: uint256
struct Bar:
    a: uint256
    b: uint256
interface IBar:
    def bar() -> Bar: view

@external
def foo(addr: address) -> Foo:
    return Foo(
        a=1,
        b=2,
        c=(staticcall IBar(addr).bar()).a,
        d=4,
        e=5
    )
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (1, 2, 3, 4, 5)


def test_external_call_in_return_single_struct(get_contract):
    code = """
struct Bar:
    a: uint256
@view
@external
def bar() -> Bar:
    return Bar(a=3)
    """

    code2 = """
struct Foo:
    a: uint256
struct Bar:
    a: uint256
interface IBar:
    def bar() -> Bar: view

@external
def foo(addr: address) -> Foo:
    return Foo(a=(staticcall IBar(addr).bar()).a)
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (3,)


def test_nested_external_call_in_return_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256

@view
@external
def bar() -> Bar:
    return Bar(a=3, b=4)

@view
@external
def baz(x: uint256) -> uint256:
    return x+1
    """

    code2 = """
struct Foo:
    a: uint256
    b: uint256
    c: uint256
    d: uint256
    e: uint256
struct Bar:
    a: uint256
    b: uint256

interface IBar:
    def bar() -> Bar: view
    def baz(a: uint256) -> uint256: view

@external
def foo(addr: address) -> Foo:
    return Foo(
        a=1,
        b=2,
        c=(staticcall IBar(addr).bar()).a,
        d=4,
        e=(staticcall IBar(addr).baz((staticcall IBar(addr).bar()).b))
    )
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (1, 2, 3, 4, 5)


def test_nested_external_call_in_return_single_struct(get_contract):
    code = """
struct Bar:
    a: uint256

@view
@external
def bar() -> Bar:
    return Bar(a=3)

@view
@external
def baz(x: uint256) -> uint256:
    return x+1
    """

    code2 = """
struct Foo:
    a: uint256
struct Bar:
    a: uint256

interface IBar:
    def bar() -> Bar: view
    def baz(a: uint256) -> uint256: view

@external
def foo(addr: address) -> Foo:
    return Foo(
        a=staticcall IBar(addr).baz((staticcall IBar(addr).bar()).a)
    )
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (4,)


@pytest.mark.parametrize("string", ["a", "abc", "abcde", "potato"])
def test_string_inside_struct(get_contract, string):
    code = f"""
struct Person:
    name: String[6]
    age: uint256

@external
def test_return() -> Person:
    return Person(name="{string}", age=42)
    """
    c1 = get_contract(code)

    code = """
struct Person:
    name: String[6]
    age: uint256

interface jsonabi:
    def test_return() -> Person: view

@external
def test_values(a: address) -> Person:
    return staticcall jsonabi(a).test_return()
    """

    c2 = get_contract(code)
    assert c2.test_values(c1.address) == (string, 42)


@pytest.mark.parametrize("string", ["a", "abc", "abcde", "potato"])
def test_string_inside_single_struct(get_contract, string):
    code = f"""
struct Person:
    name: String[6]

@external
def test_return() -> Person:
    return Person(name="{string}")
    """
    c1 = get_contract(code)

    code = """
struct Person:
    name: String[6]

interface jsonabi:
    def test_return() -> Person: view

@external
def test_values(a: address) -> Person:
    return staticcall jsonabi(a).test_return()
    """

    c2 = get_contract(code)
    assert c2.test_values(c1.address) == (string,)



def test_blockhash(get_contract):
    block_number_code = """
@external
def prev() -> bytes32:
    return block.prevhash

@external
def previous_blockhash() -> bytes32:
    return blockhash(block.number - 1)
"""
    c = get_contract(block_number_code)
    assert c.prev() == c.previous_blockhash()


def test_negative_blockhash(assert_compile_failed, get_contract):
    code = """
@external
def foo() -> bytes32:
    return blockhash(-1)
"""
    assert_compile_failed(lambda: get_contract(code))


def test_too_old_blockhash(tx_failed, get_contract, env):
    env.block_number += 257
    code = """
@external
def get_50_blockhash() -> bytes32:
    return blockhash(block.number - 257)
"""
    c = get_contract(code)
    with tx_failed():
        c.get_50_blockhash()


def test_non_existing_blockhash(tx_failed, get_contract):
    code = """
@external
def get_future_blockhash() -> bytes32:
    return blockhash(block.number + 1)
"""
    c = get_contract(code)
    with tx_failed():
        c.get_future_blockhash()



import pytest

from vyper import compiler
from vyper.exceptions import FunctionDeclarationException

fail_list = [
    """
@external
def unmatched_tupl_length() -> (Bytes[8], int128, Bytes[8]):
    return "test", 123
    """
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_tuple_return_fail(bad_code):
    with pytest.raises(FunctionDeclarationException):
        compiler.compile_code(bad_code)


def test_self_call_in_return_tuple(get_contract):
    code = """
@internal
def _foo() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 3

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return 1, 2, self._foo(), 4, 5
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_call_in_call(get_contract):
    code = """
@internal
def _foo(a: uint256, b: uint256, c: uint256) -> (uint256, uint256, uint256, uint256, uint256):
    return 1, a, b, c, 5

@internal
def _foo2() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,15,16]
    return 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return self._foo(2, 3, self._foo2())
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_nested_calls_in_tuple_return(get_contract):
    code = """
@internal
def _foo(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return 415, 3

@internal
def _foo2(a: uint256) -> uint256:
    b: uint256[10] = [6,7,8,9,10,11,12,13,14,15]
    return 99

@internal
def _foo3(a: uint256, b: uint256) -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 42

@internal
def _foo4() -> uint256:
    c: uint256[10] = [14,15,16,17,18,19,20,21,22,23]
    return 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return 1, 2, self._foo(6, 7, self._foo2(self._foo3(9, 11)))[1], self._foo4(), 5
    """

    c = get_contract(code)

    assert c.foo() == (1, 2, 3, 4, 5)


def test_external_call_in_return_tuple(get_contract):
    code = """
@view
@external
def foo() -> (uint256, uint256):
    return 3, 4
    """

    code2 = """
interface Foo:
    def foo() -> (uint256, uint256): view

@external
def foo(a: address) -> (uint256, uint256, uint256, uint256, uint256):
    return 1, 2, (staticcall Foo(a).foo())[0], 4, 5
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (1, 2, 3, 4, 5)


def test_nested_external_call_in_return_tuple(get_contract):
    code = """
@view
@external
def foo() -> (uint256, uint256):
    return 3, 4

@view
@external
def bar(a: uint256) -> uint256:
    return a+1
    """

    code2 = """
interface Foo:
    def foo() -> (uint256, uint256): view
    def bar(a: uint256) -> uint256: view

@external
def foo(a: address) -> (uint256, uint256, uint256, uint256, uint256):
    return (
        1,
        2,
        (staticcall Foo(a).foo())[0],
        4,
        staticcall Foo(a).bar((staticcall Foo(a).foo())[1])
    )
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c2.foo(c.address) == (1, 2, 3, 4, 5)


def test_single_type_tuple_int(get_contract):
    code = """
@view
@external
def foo() -> (uint256[3], uint256, uint256[2][2]):
    return [1,2,3], 4, [[5,6], [7,8]]

@view
@external
def foo2(a: int128, b: int128) -> (int128[5], int128, int128[2]):
    return [1,2,3,a,5], b, [7,8]
    """

    c = get_contract(code)

    assert c.foo() == ([1, 2, 3], 4, [[5, 6], [7, 8]])
    assert c.foo2(4, 6) == ([1, 2, 3, 4, 5], 6, [7, 8])


def test_single_type_tuple_address(get_contract):
    code = """
@view
@external
def foo() -> (address, address[2]):
    return (
        self,
        [0xF5D4020dCA6a62bB1efFcC9212AAF3c9819E30D7, 0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF]
    )
    """

    c = get_contract(code)

    assert c.foo() == (
        c.address,
        [
            "0xF5D4020dCA6a62bB1efFcC9212AAF3c9819E30D7",
            "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF",
        ],
    )


def test_single_type_tuple_bytes(get_contract):
    code = """
@view
@external
def foo() -> (Bytes[5], Bytes[5]):
    return b"hello", b"there"
    """

    c = get_contract(code)

    assert c.foo() == (b"hello", b"there")



def test_memory_deallocation(get_contract):
    code = """
event Shimmy:
    a: indexed(address)
    b: uint256

interface Other:
    def sendit(): nonpayable

@external
def foo(target: address) -> uint256[2]:
    log Shimmy(empty(address), 3)
    amount: uint256 = 1
    flargen: uint256 = 42
    extcall Other(target).sendit()
    return [amount, flargen]
    """

    code2 = """

@external
def sendit() -> bool:
    return True
    """

    c = get_contract(code)
    c2 = get_contract(code2)

    assert c.foo(c2.address) == [1, 42]



import pytest

from vyper.compiler import compile_code
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import OptimizationLevel, Settings
from vyper.ir.compile_ir import _merge_jumpdests

codes = [
    """
s: uint256

@internal
def ctor_only():
    self.s = 1

@internal
def runtime_only():
    self.s = 2

@external
def bar():
    self.runtime_only()

@deploy
def __init__():
    self.ctor_only()
    """,
    # code with nested function in it
    """
s: uint256

@internal
def runtime_only():
    self.s = 1

@internal
def foo():
    self.runtime_only()

@internal
def ctor_only():
    self.s += 1

@external
def bar():
    self.foo()

@deploy
def __init__():
    self.ctor_only()
    """,
    # code with loop in it, these are harder for dead code eliminator
    """
s: uint256

@internal
def ctor_only():
    self.s = 1

@internal
def runtime_only():
    for i: uint256 in range(10):
        self.s += 1

@external
def bar():
    self.runtime_only()

@deploy
def __init__():
    self.ctor_only()
    """,
]


# check dead code eliminator works on unreachable functions
# CMC 2024-02-05 this is not really the asm eliminator anymore,
# it happens during function code generation in module.py. so we don't
# need to test this using asm anymore.
@pytest.mark.parametrize("code", codes)
def test_dead_code_eliminator(code):
    c = CompilerData(code, settings=Settings(optimize=OptimizationLevel.NONE))

    # get the labels
    initcode_asm = [i for i in c.assembly if isinstance(i, str)]
    runtime_asm = [i for i in c.assembly_runtime if isinstance(i, str)]

    ctor_only = "ctor_only()"
    runtime_only = "runtime_only()"

    # qux reachable from unoptimized initcode, foo not reachable.
    assert any(ctor_only in instr for instr in initcode_asm)
    assert all(runtime_only not in instr for instr in initcode_asm)

    assert any(runtime_only in instr for instr in runtime_asm)
    assert all(ctor_only not in instr for instr in runtime_asm)


def test_library_code_eliminator(make_input_bundle, experimental_codegen):
    library = """
@internal
def unused1():
    pass

@internal
def unused2():
    self.unused1()

@internal
def some_function():
    pass
    """
    code = """
import library

@external
def foo():
    library.some_function()
    """
    input_bundle = make_input_bundle({"library.vy": library})
    res = compile_code(code, input_bundle=input_bundle, output_formats=["asm"])
    asm = res["asm"]
    assert "some_function()" in asm

    assert "unused1()" not in asm
    assert "unused2()" not in asm


def test_merge_jumpdests():
    asm = ["_sym_label_0", "JUMP", "PUSH0", "_sym_label_0", "JUMPDEST", "_sym_label_0", "JUMPDEST"]

    assert _merge_jumpdests(asm) is False, "should not return True as no changes were made"



import ast as python_ast
import tokenize
from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

import asttokens

from vyper.ast import nodes as vy_ast
from vyper.ast.pre_parser import pre_parse
from vyper.compiler.settings import Settings
from vyper.exceptions import CompilerPanic, ParserException, SyntaxException
from vyper.typing import ModificationOffsets
from vyper.utils import sha256sum, vyper_warn


def parse_to_ast(*args: Any, **kwargs: Any) -> vy_ast.Module:
    _settings, ast = parse_to_ast_with_settings(*args, **kwargs)
    return ast


def parse_to_ast_with_settings(
    vyper_source: str,
    source_id: int = 0,
    module_path: Optional[str] = None,
    resolved_path: Optional[str] = None,
    add_fn_node: Optional[str] = None,
) -> tuple[Settings, vy_ast.Module]:
    """
    Parses a Vyper source string and generates basic Vyper AST nodes.

    Parameters
    ----------
    vyper_source: str
        The Vyper source code to parse.
    source_id : int, optional
        Source id to use in the `src` member of each node.
    contract_name: str, optional
        Name of contract.
    add_fn_node: str, optional
        If not None, adds a dummy Python AST FunctionDef wrapper node.
    source_id: int, optional
        The source ID generated for this source code.
        Corresponds to FileInput.source_id
    module_path: str, optional
        The path of the source code
        Corresponds to FileInput.path
    resolved_path: str, optional
        The resolved path of the source code
        Corresponds to FileInput.resolved_path

    Returns
    -------
    list
        Untyped, unoptimized Vyper AST nodes.
    """
    if "\x00" in vyper_source:
        raise ParserException("No null bytes (\\x00) allowed in the source code.")
    settings, class_types, for_loop_annotations, python_source = pre_parse(vyper_source)
    try:
        py_ast = python_ast.parse(python_source)
    except SyntaxError as e:
        # TODO: Ensure 1-to-1 match of source_code:reformatted_code SyntaxErrors
        raise SyntaxException(str(e), vyper_source, e.lineno, e.offset) from None

    # Add dummy function node to ensure local variables are treated as `AnnAssign`
    # instead of state variables (`VariableDecl`)
    if add_fn_node:
        fn_node = python_ast.FunctionDef(add_fn_node, py_ast.body, [], [])
        fn_node.body = py_ast.body
        fn_node.args = python_ast.arguments(defaults=[])
        py_ast.body = [fn_node]

    annotate_python_ast(
        py_ast,
        vyper_source,
        class_types,
        for_loop_annotations,
        source_id=source_id,
        module_path=module_path,
        resolved_path=resolved_path,
    )

    # postcondition: consumed all the for loop annotations
    assert len(for_loop_annotations) == 0

    # Convert to Vyper AST.
    module = vy_ast.get_node(py_ast)
    assert isinstance(module, vy_ast.Module)  # mypy hint

    return settings, module


def ast_to_dict(ast_struct: Union[vy_ast.VyperNode, List]) -> Union[Dict, List]:
    """
    Converts a Vyper AST node, or list of nodes, into a dictionary suitable for
    output to the user.
    """
    if isinstance(ast_struct, vy_ast.VyperNode):
        return ast_struct.to_dict()

    if isinstance(ast_struct, list):
        return [i.to_dict() for i in ast_struct]

    raise CompilerPanic(f'Unknown Vyper AST node provided: "{type(ast_struct)}".')


def dict_to_ast(ast_struct: Union[Dict, List]) -> Union[vy_ast.VyperNode, List]:
    """
    Converts an AST dict, or list of dicts, into Vyper AST node objects.
    """
    if isinstance(ast_struct, dict):
        return vy_ast.get_node(ast_struct)
    if isinstance(ast_struct, list):
        return [vy_ast.get_node(i) for i in ast_struct]
    raise CompilerPanic(f'Unknown ast_struct provided: "{type(ast_struct)}".')


def annotate_python_ast(
    parsed_ast: python_ast.AST,
    vyper_source: str,
    modification_offsets: ModificationOffsets,
    for_loop_annotations: dict,
    source_id: int = 0,
    module_path: Optional[str] = None,
    resolved_path: Optional[str] = None,
) -> python_ast.AST:
    """
    Annotate and optimize a Python AST in preparation conversion to a Vyper AST.

    Parameters
    ----------
    parsed_ast : AST
        The AST to be annotated and optimized.
    vyper_source: str
        The original vyper source code
    loop_var_annotations: dict
        A mapping of line numbers of `For` nodes to the tokens of the type
        annotation of the iterator extracted during pre-parsing.
    modification_offsets : dict
        A mapping of class names to their original class types.

    Returns
    -------
        The annotated and optimized AST.
    """
    tokens = asttokens.ASTTokens(vyper_source)
    assert isinstance(parsed_ast, python_ast.Module)  # help mypy
    tokens.mark_tokens(parsed_ast)
    visitor = AnnotatingVisitor(
        vyper_source,
        modification_offsets,
        for_loop_annotations,
        tokens,
        source_id,
        module_path=module_path,
        resolved_path=resolved_path,
    )
    visitor.visit(parsed_ast)

    return parsed_ast


class AnnotatingVisitor(python_ast.NodeTransformer):
    _source_code: str
    _modification_offsets: ModificationOffsets
    _loop_var_annotations: dict[int, dict[str, Any]]

    def __init__(
        self,
        source_code: str,
        modification_offsets: ModificationOffsets,
        for_loop_annotations: dict,
        tokens: asttokens.ASTTokens,
        source_id: int,
        module_path: Optional[str] = None,
        resolved_path: Optional[str] = None,
    ):
        self._tokens = tokens
        self._source_id = source_id
        self._module_path = module_path
        self._resolved_path = resolved_path
        self._source_code = source_code
        self._modification_offsets = modification_offsets
        self._for_loop_annotations = for_loop_annotations

        self.counter: int = 0

    def generic_visit(self, node):
        """
        Annotate a node with information that simplifies Vyper node generation.
        """
        # Decorate every node with the original source code to allow pretty-printing errors
        node.full_source_code = self._source_code
        node.node_id = self.counter
        node.ast_type = node.__class__.__name__
        self.counter += 1

        # Decorate every node with source end offsets
        start = (None, None)
        if hasattr(node, "first_token"):
            start = node.first_token.start
        end = (None, None)
        if hasattr(node, "last_token"):
            end = node.last_token.end
            if node.last_token.type == 4:
                # token type 4 is a `\n`, some nodes include a trailing newline
                # here we ignore it when building the node offsets
                end = (end[0], end[1] - 1)

        node.lineno = start[0]
        node.col_offset = start[1]
        node.end_lineno = end[0]
        node.end_col_offset = end[1]

        # TODO: adjust end_lineno and end_col_offset when this node is in
        # modification_offsets

        if hasattr(node, "last_token"):
            start_pos = node.first_token.startpos
            end_pos = node.last_token.endpos

            if node.last_token.type == 4:
                # ignore trailing newline once more
                end_pos -= 1
            node.src = f"{start_pos}:{end_pos-start_pos}:{self._source_id}"
            node.node_source_code = self._source_code[start_pos:end_pos]

        return super().generic_visit(node)

    def _visit_docstring(self, node):
        """
        Move a node docstring from body to `doc_string` and annotate it as `DocStr`.
        """
        self.generic_visit(node)

        if node.body:
            n = node.body[0]
            if (
                isinstance(n, python_ast.Expr)
                and isinstance(n.value, python_ast.Constant)
                and isinstance(n.value.value, str)
            ):
                self.generic_visit(n.value)
                n.value.ast_type = "DocStr"
                del node.body[0]
                node.doc_string = n.value

        return node

    def visit_Module(self, node):
        # TODO: is this the best place for these? maybe they can be on
        # CompilerData instead.
        node.path = self._module_path
        node.resolved_path = self._resolved_path
        node.source_sha256sum = sha256sum(self._source_code)
        node.source_id = self._source_id
        return self._visit_docstring(node)

    def visit_FunctionDef(self, node):
        if node.decorator_list:
            # start the source highlight at `def` to improve annotation readability
            decorator_token = node.decorator_list[-1].last_token
            def_token = self._tokens.find_token(decorator_token, tokenize.NAME, tok_str="def")
            node.first_token = def_token

        return self._visit_docstring(node)

    def visit_ClassDef(self, node):
        """
        Convert the `ClassDef` node into a Vyper-specific node type.

        Vyper uses `struct` and `interface` in place of `class`, however these
        values must be substituted out to create parseable Python. The Python
        node is annotated with the desired Vyper type via the `ast_type` member.
        """
        self.generic_visit(node)

        node.ast_type = self._modification_offsets[(node.lineno, node.col_offset)]
        return node

    def visit_For(self, node):
        """
        Visit a For node, splicing in the loop variable annotation provided by
        the pre-parser
        """
        annotation_tokens = self._for_loop_annotations.pop((node.lineno, node.col_offset))

        if not annotation_tokens:
            # a common case for people migrating to 0.4.0, provide a more
            # specific error message than "invalid type annotation"
            raise SyntaxException(
                "missing type annotation\n\n"
                "  (hint: did you mean something like "
                f"`for {node.target.id}: uint256 in ...`?)",
                self._source_code,
                node.lineno,
                node.col_offset,
            )

        # some kind of black magic. untokenize preserves the line and column
        # offsets, giving us something like `\
        # \
        # \
        #   uint8`
        # that's not a valid python Expr because it is indented.
        # but it's good because the code is indented to exactly the same
        # offset as it did in the original source!
        # (to best understand this, print out annotation_str and
        # self._source_code and compare them side-by-side).
        #
        # what we do here is add in a dummy target which we will remove
        # in a bit, but for now lets us keep the line/col offset, and
        # *also* gives us a valid AST. it doesn't matter what the dummy
        # target name is, since it gets removed in a few lines.
        annotation_str = tokenize.untokenize(annotation_tokens)
        annotation_str = "dummy_target:" + annotation_str

        try:
            fake_node = python_ast.parse(annotation_str).body[0]
        except SyntaxError as e:
            raise SyntaxException(
                "invalid type annotation", self._source_code, node.lineno, node.col_offset
            ) from e

        # fill in with asttokens info. note we can use `self._tokens` because
        # it is indented to exactly the same position where it appeared
        # in the original source!
        self._tokens.mark_tokens(fake_node)

        # replace the dummy target name with the real target name.
        fake_node.target = node.target
        # replace the For node target with the new ann_assign
        node.target = fake_node

        return self.generic_visit(node)

    def visit_Expr(self, node):
        """
        Convert the `Yield` node into a Vyper-specific node type.

        Vyper substitutes `yield` for non-pythonic statement such as `log`. Prior
        to generating Vyper AST, we must annotate `Yield` nodes with their original
        value.

        Because `Yield` is an expression-statement, we also remove it from it's
        enclosing `Expr` node.
        """
        self.generic_visit(node)

        if isinstance(node.value, python_ast.Yield):
            # CMC 2024-03-03 consider unremoving this from the enclosing Expr
            node = node.value
            node.ast_type = self._modification_offsets[(node.lineno, node.col_offset)]

        return node

    def visit_Await(self, node):
        start_pos = node.lineno, node.col_offset  # grab these before generic_visit modifies them
        self.generic_visit(node)
        node.ast_type = self._modification_offsets[start_pos]
        return node

    def visit_Call(self, node):
        # Convert structs declared as `Dict` node for vyper < 0.4.0 to kwargs
        if len(node.args) == 1 and isinstance(node.args[0], python_ast.Dict):
            msg = "Instantiating a struct using a dictionary is deprecated "
            msg += "as of v0.4.0 and will be disallowed in a future release. "
            msg += "Use kwargs instead e.g. Foo(a=1, b=2)"

            # add full_source_code so that str(VyperException(msg, node)) works
            node.full_source_code = self._source_code
            vyper_warn(msg, node)

            dict_ = node.args[0]
            kw_list = []

            assert len(dict_.keys) == len(dict_.values)
            for key, value in zip(dict_.keys, dict_.values):
                replacement_kw_node = python_ast.keyword(key.id, value)
                kw_list.append(replacement_kw_node)

            node.args = []
            node.keywords = kw_list

        self.generic_visit(node)

        return node

    def visit_Constant(self, node):
        """
        Handle `Constant` when using Python >=3.8

        In Python 3.8, `NameConstant`, `Num`, `Str`, and `Bytes` are deprecated
        in favor of `Constant`. To maintain consistency across versions, `ast_type`
        is modified to create the <=3.7 node classes.
        """
        if not isinstance(node.value, bool) and isinstance(node.value, (int, float)):
            return self.visit_Num(node)

        self.generic_visit(node)
        if node.value is None or isinstance(node.value, bool):
            node.ast_type = "NameConstant"
        elif isinstance(node.value, str):
            node.ast_type = "Str"
        elif isinstance(node.value, bytes):
            node.ast_type = "Bytes"
        elif isinstance(node.value, Ellipsis.__class__):
            node.ast_type = "Ellipsis"
        else:
            raise SyntaxException(
                "Invalid syntax (unsupported Python Constant AST node).",
                self._source_code,
                node.lineno,
                node.col_offset,
            )

        return node

    def visit_Num(self, node):
        """
        Adjust numeric node class based on the value type.

        Python uses `Num` to represent floats and integers. Integers may also
        be given in binary, octal, decimal, or hexadecimal format. This method
        modifies `ast_type` to separate `Num` into more granular Vyper node
        classes.
        """
        # modify vyper AST type according to the format of the literal value
        self.generic_visit(node)
        value = node.node_source_code

        # deduce non base-10 types based on prefix
        if value.lower()[:2] == "0x":
            if len(value) % 2:
                raise SyntaxException(
                    "Hex notation requires an even number of digits",
                    self._source_code,
                    node.lineno,
                    node.col_offset,
                )
            node.ast_type = "Hex"
            node.value = value

        elif value.lower()[:2] == "0b":
            node.ast_type = "Bytes"
            mod = (len(value) - 2) % 8
            if mod:
                raise SyntaxException(
                    f"Bit notation requires a multiple of 8 bits. {8-mod} bit(s) are missing.",
                    self._source_code,
                    node.lineno,
                    node.col_offset,
                )
            node.value = int(value, 2).to_bytes(len(value) // 8, "big")

        elif isinstance(node.value, float):
            node.ast_type = "Decimal"
            node.value = Decimal(value)

        elif isinstance(node.value, int):
            node.ast_type = "Int"

        else:  # pragma: nocover
            raise CompilerPanic(f"Unexpected type for Constant value: {type(node.value).__name__}")

        return node

    def visit_UnaryOp(self, node):
        """
        Adjust operand value and discard unary operations, where possible.

        This is done so that negative decimal literals are accurately represented.
        """
        self.generic_visit(node)

        is_sub = isinstance(node.op, python_ast.USub)
        is_num = hasattr(node.operand, "value") and isinstance(node.operand.value, (int, Decimal))
        if is_sub and is_num:
            node.operand.value = 0 - node.operand.value
            node.operand.col_offset = node.col_offset
            node.operand.node_source_code = node.node_source_code
            return node.operand
        else:
            return node



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import ArgumentException, SyntaxException

fail_list = [
    """
@external
def foo():
    x: address = create_minimal_proxy_to(
        0x1234567890123456789012345678901234567890,
        value=4,
        value=9
    )
    """,
    """
@external
def foo(_salt: bytes32):
    x: address = create_minimal_proxy_to(
        0x1234567890123456789012345678901234567890, salt=keccak256(b"Vyper Rocks!"), salt=_salt
    )
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_type_mismatch_exception(bad_code):
    with raises((SyntaxException, ArgumentException)):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    # test that create_forwarder_to is valid syntax; the name is deprecated
    x: address = create_forwarder_to(0x1234567890123456789012345678901234567890)
    """,
    """
@external
def foo():
    x: address = create_minimal_proxy_to(0x1234567890123456789012345678901234567890)
    """,
    """
@external
def foo():
    x: address = create_minimal_proxy_to(
        0x1234567890123456789012345678901234567890,
        value=as_wei_value(9, "wei")
    )
    """,
    """
@external
def foo():
    x: address = create_minimal_proxy_to(0x1234567890123456789012345678901234567890, value=9)
    """,
    """
@external
def foo():
    x: address = create_minimal_proxy_to(
        0x1234567890123456789012345678901234567890,
        salt=keccak256(b"Vyper Rocks!")
    )
    """,
    """
@external
def foo(_salt: bytes32):
    x: address = create_minimal_proxy_to(0x1234567890123456789012345678901234567890, salt=_salt)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_rlp_success(good_code):
    assert compiler.compile_code(good_code) is not None



# maybe rename this `main.py` or `venom.py`
# (can have an `__init__.py` which exposes the API).

from typing import Optional

from vyper.codegen.ir_node import IRnode
from vyper.compiler.settings import OptimizationLevel
from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.context import IRContext
from vyper.venom.function import IRFunction
from vyper.venom.ir_node_to_venom import ir_node_to_venom
from vyper.venom.passes.dft import DFTPass
from vyper.venom.passes.make_ssa import MakeSSA
from vyper.venom.passes.mem2var import Mem2Var
from vyper.venom.passes.remove_unused_variables import RemoveUnusedVariablesPass
from vyper.venom.passes.sccp import SCCP
from vyper.venom.passes.simplify_cfg import SimplifyCFGPass
from vyper.venom.passes.store_elimination import StoreElimination
from vyper.venom.venom_to_assembly import VenomCompiler

DEFAULT_OPT_LEVEL = OptimizationLevel.default()


def generate_assembly_experimental(
    runtime_code: IRContext,
    deploy_code: Optional[IRContext] = None,
    optimize: OptimizationLevel = DEFAULT_OPT_LEVEL,
) -> list[str]:
    # note: VenomCompiler is sensitive to the order of these!
    if deploy_code is not None:
        functions = [deploy_code, runtime_code]
    else:
        functions = [runtime_code]

    compiler = VenomCompiler(functions)
    return compiler.generate_evm(optimize == OptimizationLevel.NONE)


def _run_passes(fn: IRFunction, optimize: OptimizationLevel) -> None:
    # Run passes on Venom IR
    # TODO: Add support for optimization levels

    ac = IRAnalysesCache(fn)

    SimplifyCFGPass(ac, fn).run_pass()
    Mem2Var(ac, fn).run_pass()
    MakeSSA(ac, fn).run_pass()
    StoreElimination(ac, fn).run_pass()
    MakeSSA(ac, fn).run_pass()
    SCCP(ac, fn).run_pass()
    SimplifyCFGPass(ac, fn).run_pass()
    RemoveUnusedVariablesPass(ac, fn).run_pass()
    DFTPass(ac, fn).run_pass()


def generate_ir(ir: IRnode, optimize: OptimizationLevel) -> IRContext:
    # Convert "old" IR to "new" IR
    ctx = ir_node_to_venom(ir)
    for fn in ctx.functions.values():
        _run_passes(fn, optimize)

    return ctx



import pytest

from vyper import compiler

valid_list = [
    """
x: public(int128)
    """,
    """
x: public(constant(int128)) = 0
y: public(immutable(int128))

@deploy
def __init__():
    y = 0
    """,
    """
x: public(int128)
y: public(int128)
z: public(int128)

@external
def foo() -> int128:
    return self.x // self.y // self.z
    """,
    # expansion of public user-defined struct
    """
struct Foo:
    a: uint256

x: public(HashMap[uint256, Foo])
    """,
    # expansion of public user-defined flag
    """
flag Foo:
    BAR

x: public(HashMap[uint256, Foo])
    """,
    # expansion of public user-defined interface
    """
interface Foo:
    def bar(): nonpayable

x: public(HashMap[uint256, Foo])
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_public_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest
from hexbytes import HexBytes

from tests.utils import ZERO_ADDRESS
from vyper import compile_code
from vyper.builtins.functions import eip1167_bytecode
from vyper.exceptions import ArgumentException, StateAccessViolation, TypeMismatch


def test_max_outsize_exceeds_returndatasize(get_contract):
    source_code = """
@external
def foo() -> Bytes[7]:
    return raw_call(0x0000000000000000000000000000000000000004, b"moose", max_outsize=7)
    """
    c = get_contract(source_code)
    assert c.foo() == b"moose"


def test_raw_call_non_memory(get_contract):
    source_code = """
_foo: Bytes[5]
@external
def foo() -> Bytes[5]:
    self._foo = b"moose"
    return raw_call(0x0000000000000000000000000000000000000004, self._foo, max_outsize=5)
    """
    c = get_contract(source_code)
    assert c.foo() == b"moose"


def test_returndatasize_exceeds_max_outsize(get_contract):
    source_code = """
@external
def foo() -> Bytes[3]:
    return raw_call(0x0000000000000000000000000000000000000004, b"moose", max_outsize=3)
    """
    c = get_contract(source_code)
    assert c.foo() == b"moo"


def test_returndatasize_matches_max_outsize(get_contract):
    source_code = """
@external
def foo() -> Bytes[5]:
    return raw_call(0x0000000000000000000000000000000000000004, b"moose", max_outsize=5)
    """
    c = get_contract(source_code)
    assert c.foo() == b"moose"


def test_multiple_levels(env, get_contract):
    inner_code = """
@external
def returnten() -> int128:
    return 10
    """

    c = get_contract(inner_code)

    outer_code = """
@external
def create_and_call_returnten(inp: address) -> int128:
    x: address = create_minimal_proxy_to(inp)
    o: int128 = extract32(raw_call(x, b"\\xd0\\x1f\\xb1\\xb8", max_outsize=32, gas=50000), 0, output_type=int128)  # noqa: E501
    return o

@external
def create_and_return_proxy(inp: address) -> address:
    x: address = create_minimal_proxy_to(inp)
    return x
    """

    c2 = get_contract(outer_code)
    assert c2.create_and_call_returnten(c.address) == 10
    c2.create_and_call_returnten(c.address)

    _, preamble, callcode = eip1167_bytecode()

    c3 = c2.create_and_return_proxy(c.address)
    c2.create_and_return_proxy(c.address)

    c3_contract_code = env.get_code(c3)

    assert c3_contract_code[:10] == HexBytes(preamble)
    assert c3_contract_code[-15:] == HexBytes(callcode)

    print("Passed proxy test")
    # TODO: This one is special
    # print(f'Gas consumed: {(chain.head_state.receipts[-1].gas_used - chain.head_state.receipts[-2].gas_used - chain.last_tx.intrinsic_gas_used)}')  # noqa: E501


def test_multiple_levels2(tx_failed, get_contract):
    inner_code = """
@external
def returnten() -> int128:
    raise
    """

    c = get_contract(inner_code)

    outer_code = """
@external
def create_and_call_returnten(inp: address) -> int128:
    x: address = create_minimal_proxy_to(inp)
    o: int128 = extract32(raw_call(x, b"\\xd0\\x1f\\xb1\\xb8", max_outsize=32, gas=50000), 0, output_type=int128)  # noqa: E501
    return o

@external
def create_and_return_proxy(inp: address) -> address:
    return create_minimal_proxy_to(inp)
    """

    c2 = get_contract(outer_code)

    with tx_failed():
        c2.create_and_call_returnten(c.address)

    print("Passed minimal proxy exception test")


def test_delegate_call(env, get_contract):
    inner_code = """
a: address  # this is required for storage alignment...
owners: public(address[5])

@external
def set_owner(i: int128, o: address):
    self.owners[i] = o
    """

    inner_contract = get_contract(inner_code)

    outer_code = """
owner_setter_contract: public(address)
owners: public(address[5])


@deploy
def __init__(_owner_setter: address):
    self.owner_setter_contract = _owner_setter


@external
def set(i: int128, owner: address):
    # delegate setting owners to other contract.s
    cdata: Bytes[68] = concat(method_id("set_owner(int128,address)"), convert(i, bytes32), convert(owner, bytes32))  # noqa: E501
    raw_call(
        self.owner_setter_contract,
        cdata,
        gas=msg.gas,
        max_outsize=0,
        is_delegate_call=True
    )
    """

    a0, a1, a2 = env.accounts[:3]
    outer_contract = get_contract(outer_code, inner_contract.address)

    # Test setting on inners contract's state setting works.
    inner_contract.set_owner(1, a2)
    assert inner_contract.owners(1) == a2

    # Confirm outer contract's state is empty and contract to call has been set.
    assert outer_contract.owner_setter_contract() == inner_contract.address
    assert outer_contract.owners(1) == ZERO_ADDRESS

    # Call outer contract, that make a delegate call to inner_contract.
    outer_contract.set(1, a1)
    assert outer_contract.owners(1) == a1


def test_gas(get_contract, tx_failed):
    inner_code = """
bar: bytes32

@external
def foo(_bar: bytes32):
    self.bar = _bar
    """

    inner_contract = get_contract(inner_code)

    outer_code = """
@external
def foo_call(_addr: address):
    cdata: Bytes[40] = concat(
        method_id("foo(bytes32)"),
        0x0000000000000000000000000000000000000000000000000000000000000001
    )
    raw_call(_addr, cdata, max_outsize=0{})
    """

    # with no gas value given, enough will be forwarded to complete the call
    outer_contract = get_contract(outer_code.format(""))
    outer_contract.foo_call(inner_contract.address)

    # manually specifying a sufficient amount should succeed
    outer_contract = get_contract(outer_code.format(", gas=50000"))
    outer_contract.foo_call(inner_contract.address)

    # manually specifying an insufficient amount should fail
    outer_contract = get_contract(outer_code.format(", gas=2250"))
    with tx_failed():
        outer_contract.foo_call(inner_contract.address)


def test_static_call(get_contract):
    target_source = """
@external
@view
def foo() -> int128:
    return 42
"""

    caller_source = """
@external
@view
def foo(_addr: address) -> int128:
    _response: Bytes[32] = raw_call(
        _addr,
        method_id("foo()"),
        max_outsize=32,
        is_static_call=True,
    )
    return convert(_response, int128)
    """

    target = get_contract(target_source)
    caller = get_contract(caller_source)

    assert caller.foo(target.address) == 42


def test_forward_calldata(get_contract, env, keccak):
    target_source = """
@external
def foo() -> uint256:
    return 123
    """

    caller_source = """
target: address

@external
def set_target(target: address):
    self.target = target

@external
def __default__():
    assert 123 == _abi_decode(raw_call(self.target, msg.data, max_outsize=32), uint256)
    """

    target = get_contract(target_source)

    caller = get_contract(caller_source)
    caller.set_target(target.address)

    # manually construct msg.data for `caller` contract
    sig = keccak("foo()".encode()).hex()[:10]
    assert env.message_call(caller.address, data=sig) == b""


# check max_outsize=0 does same thing as not setting max_outsize.
# compile to bytecode and compare bytecode directly.
def test_max_outsize_0():
    code1 = """
@external
def test_raw_call(_target: address):
    raw_call(_target, method_id("foo()"))
    """
    code2 = """
@external
def test_raw_call(_target: address):
    raw_call(_target, method_id("foo()"), max_outsize=0)
    """
    output1 = compile_code(code1, output_formats=["bytecode", "bytecode_runtime"])
    output2 = compile_code(code2, output_formats=["bytecode", "bytecode_runtime"])
    assert output1 == output2


# check max_outsize=0 does same thing as not setting max_outsize,
# this time with revert_on_failure set to False
def test_max_outsize_0_no_revert_on_failure():
    code1 = """
@external
def test_raw_call(_target: address) -> bool:
    # compile raw_call both ways, with revert_on_failure
    a: bool = raw_call(_target, method_id("foo()"), revert_on_failure=False)
    return a
    """
    # same code, but with max_outsize=0
    code2 = """
@external
def test_raw_call(_target: address) -> bool:
    a: bool = raw_call(_target, method_id("foo()"), max_outsize=0, revert_on_failure=False)
    return a
    """
    output1 = compile_code(code1, output_formats=["bytecode", "bytecode_runtime"])
    output2 = compile_code(code2, output_formats=["bytecode", "bytecode_runtime"])
    assert output1 == output2


# test functionality of max_outsize=0
def test_max_outsize_0_call(get_contract):
    target_source = """
@external
@payable
def bar() -> uint256:
    return 123
    """

    caller_source = """
@external
@payable
def foo(_addr: address) -> bool:
    success: bool = raw_call(_addr, method_id("bar()"), max_outsize=0, revert_on_failure=False)
    return success
    """

    target = get_contract(target_source)
    caller = get_contract(caller_source)
    assert caller.foo(target.address) is True


def test_static_call_fails_nonpayable(get_contract, tx_failed):
    target_source = """
baz: int128

@external
def foo() -> int128:
    self.baz = 31337
    return self.baz
"""

    caller_source = """
@external
@view
def foo(_addr: address) -> int128:
    _response: Bytes[32] = raw_call(
        _addr,
        method_id("foo()"),
        max_outsize=32,
        is_static_call=True,
    )
    return convert(_response, int128)
    """

    target = get_contract(target_source)
    caller = get_contract(caller_source)

    with tx_failed():
        caller.foo(target.address)


def test_checkable_raw_call(get_contract, tx_failed):
    target_source = """
baz: int128
@external
def fail1(should_raise: bool):
    if should_raise:
        raise "fail"

# test both paths for raw_call -
# they are different depending if callee has or doesn't have returntype
# (fail2 fails because of staticcall)
@external
def fail2(should_raise: bool) -> int128:
    if should_raise:
        self.baz = self.baz + 1
    return self.baz
"""

    caller_source = """
@external
@view
def foo(_addr: address, should_raise: bool) -> uint256:
    success: bool = True
    response: Bytes[32] = b""
    success, response = raw_call(
        _addr,
        _abi_encode(should_raise, method_id=method_id("fail1(bool)")),
        max_outsize=32,
        is_static_call=True,
        revert_on_failure=False,
    )
    assert success == (not should_raise)
    return 1

@external
@view
def bar(_addr: address, should_raise: bool) -> uint256:
    success: bool = True
    response: Bytes[32] = b""
    success, response = raw_call(
        _addr,
        _abi_encode(should_raise, method_id=method_id("fail2(bool)")),
        max_outsize=32,
        is_static_call=True,
        revert_on_failure=False,
    )
    assert success == (not should_raise)
    return 2

# test max_outsize not set case
@external
@nonpayable
def baz(_addr: address, should_raise: bool) -> uint256:
    success: bool = True
    success = raw_call(
        _addr,
        _abi_encode(should_raise, method_id=method_id("fail1(bool)")),
        revert_on_failure=False,
    )
    assert success == (not should_raise)
    return 3
    """

    target = get_contract(target_source)
    caller = get_contract(caller_source)

    assert caller.foo(target.address, True) == 1
    assert caller.foo(target.address, False) == 1
    assert caller.bar(target.address, True) == 2
    assert caller.bar(target.address, False) == 2
    assert caller.baz(target.address, True) == 3
    assert caller.baz(target.address, False) == 3


# XXX: these test_raw_call_clean_mem* tests depend on variables and
# calling convention writing to memory. think of ways to make more
# robust to changes to calling convention and memory layout.


def test_raw_call_msg_data_clean_mem(get_contract):
    # test msize uses clean memory and does not get overwritten by
    # any raw_call() arguments
    code = """
identity: constant(address) = 0x0000000000000000000000000000000000000004

@external
def foo():
    pass

@internal
@view
def get_address()->address:
    a:uint256 = 121 # 0x79
    return identity
@external
def bar(f: uint256, u: uint256) -> Bytes[100]:
    # embed an internal call in the calculation of address
    a: Bytes[100] = raw_call(self.get_address(), msg.data, max_outsize=100)
    return a
    """

    c = get_contract(code)
    assert (
        c.bar(1, 2).hex() == "ae42e951"
        "0000000000000000000000000000000000000000000000000000000000000001"
        "0000000000000000000000000000000000000000000000000000000000000002"
    )


def test_raw_call_clean_mem2(get_contract):
    # test msize uses clean memory and does not get overwritten by
    # any raw_call() arguments, another way
    code = """
buf: Bytes[100]

@external
def bar(f: uint256, g: uint256, h: uint256) -> Bytes[100]:
    # embed a memory modifying expression in the calculation of address
    self.buf = raw_call(
        [0x0000000000000000000000000000000000000004,][f-1],
        msg.data,
        max_outsize=100
    )
    return self.buf
    """
    c = get_contract(code)

    assert (
        c.bar(1, 2, 3).hex() == "9309b76e"
        "0000000000000000000000000000000000000000000000000000000000000001"
        "0000000000000000000000000000000000000000000000000000000000000002"
        "0000000000000000000000000000000000000000000000000000000000000003"
    )


def test_raw_call_clean_mem3(get_contract):
    # test msize uses clean memory and does not get overwritten by
    # any raw_call() arguments, and also test order of evaluation for
    # scope_multi
    code = """
buf: Bytes[100]
canary: String[32]

@internal
def bar() -> address:
    self.canary = "bar"
    return 0x0000000000000000000000000000000000000004

@internal
def goo() -> uint256:
    self.canary = "goo"
    return 0

@external
def foo() -> String[32]:
    self.buf = raw_call(self.bar(), msg.data, value = self.goo(), max_outsize=100)
    return self.canary
    """
    c = get_contract(code)
    assert c.foo() == "goo"


def test_raw_call_clean_mem_kwargs_value(get_contract, env):
    # test msize uses clean memory and does not get overwritten by
    # any raw_call() kwargs
    code = """
buf: Bytes[100]

# add a dummy function to trigger memory expansion in the selector table routine
@external
def foo():
    pass

@internal
def _value() -> uint256:
    x: uint256 = 1
    return x

@external
def bar(f: uint256) -> Bytes[100]:
    # embed a memory modifying expression in the calculation of address
    self.buf = raw_call(
        0x0000000000000000000000000000000000000004,
        msg.data,
        max_outsize=100,
        value=self._value()
    )
    return self.buf
    """
    env.set_balance(env.deployer, 1)
    c = get_contract(code, value=1)

    assert (
        c.bar(13).hex() == "0423a132"
        "000000000000000000000000000000000000000000000000000000000000000d"
    )


def test_raw_call_clean_mem_kwargs_gas(get_contract, env):
    # test msize uses clean memory and does not get overwritten by
    # any raw_call() kwargs
    code = """
buf: Bytes[100]

# add a dummy function to trigger memory expansion in the selector table routine
@external
def foo():
    pass

@internal
def _gas() -> uint256:
    x: uint256 = msg.gas
    return x

@external
def bar(f: uint256) -> Bytes[100]:
    # embed a memory modifying expression in the calculation of address
    self.buf = raw_call(
        0x0000000000000000000000000000000000000004,
        msg.data,
        max_outsize=100,
        gas=self._gas()
    )
    return self.buf
    """
    env.set_balance(env.deployer, 1)
    c = get_contract(code, value=1)

    assert (
        c.bar(15).hex() == "0423a132"
        "000000000000000000000000000000000000000000000000000000000000000f"
    )


uncompilable_code = [
    (
        """
@external
@view
def foo(_addr: address):
    raw_call(_addr, method_id("foo()"))
    """,
        StateAccessViolation,
    ),
    (
        """
@external
def foo(_addr: address):
    raw_call(_addr, method_id("foo()"), is_delegate_call=True, is_static_call=True)
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo(_addr: address):
    raw_call(_addr, method_id("foo()"), is_delegate_call=True, value=1)
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo(_addr: address):
    raw_call(_addr, method_id("foo()"), is_static_call=True, value=1)
    """,
        ArgumentException,
    ),
    (
        """
@external
@view
def foo(_addr: address):
    raw_call(_addr, 256)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("source_code,exc", uncompilable_code)
def test_invalid_type_exception(assert_compile_failed, get_contract, source_code, exc):
    assert_compile_failed(lambda: get_contract(source_code), exc)



from typing import TYPE_CHECKING, Any, Iterator, Optional, Union

from vyper.codegen.ir_node import IRnode
from vyper.utils import OrderedSet

# instructions which can terminate a basic block
BB_TERMINATORS = frozenset(["jmp", "djmp", "jnz", "ret", "return", "stop", "exit"])

VOLATILE_INSTRUCTIONS = frozenset(
    [
        "param",
        "call",
        "staticcall",
        "delegatecall",
        "invoke",
        "sload",
        "sstore",
        "iload",
        "istore",
        "tload",
        "tstore",
        "assert",
        "assert_unreachable",
        "mstore",
        "mload",
        "calldatacopy",
        "mcopy",
        "extcodecopy",
        "returndatacopy",
        "codecopy",
        "dloadbytes",
        "dload",
        "return",
        "ret",
        "jmp",
        "jnz",
    ]
)

NO_OUTPUT_INSTRUCTIONS = frozenset(
    [
        "mstore",
        "sstore",
        "dstore",
        "istore",
        "dloadbytes",
        "calldatacopy",
        "mcopy",
        "returndatacopy",
        "codecopy",
        "extcodecopy",
        "return",
        "ret",
        "revert",
        "assert",
        "assert_unreachable",
        "selfdestruct",
        "stop",
        "invalid",
        "invoke",
        "jmp",
        "djmp",
        "jnz",
        "log",
        "exit",
    ]
)

CFG_ALTERING_INSTRUCTIONS = frozenset(["jmp", "djmp", "jnz"])

if TYPE_CHECKING:
    from vyper.venom.function import IRFunction


class IRDebugInfo:
    """
    IRDebugInfo represents debug information in IR, used to annotate IR
    instructions with source code information when printing IR.
    """

    line_no: int
    src: str

    def __init__(self, line_no: int, src: str) -> None:
        self.line_no = line_no
        self.src = src

    def __repr__(self) -> str:
        src = self.src if self.src else ""
        return f"\t# line {self.line_no}: {src}".expandtabs(20)


class IROperand:
    """
    IROperand represents an IR operand. An operand is anything that can be
    operated by instructions. It can be a literal, a variable, or a label.
    """

    value: Any

    @property
    def name(self) -> str:
        return self.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.value == other.value

    def __repr__(self) -> str:
        return str(self.value)


class IRLiteral(IROperand):
    """
    IRLiteral represents a literal in IR
    """

    value: int

    def __init__(self, value: int) -> None:
        assert isinstance(value, int), "value must be an int"
        self.value = value


class IRVariable(IROperand):
    """
    IRVariable represents a variable in IR. A variable is a string that starts with a %.
    """

    value: str

    def __init__(self, value: str, version: Optional[str | int] = None) -> None:
        assert isinstance(value, str)
        assert ":" not in value, "Variable name cannot contain ':'"
        if version:
            assert isinstance(value, str) or isinstance(value, int), "value must be an str or int"
            value = f"{value}:{version}"
        if value[0] != "%":
            value = f"%{value}"
        self.value = value

    @property
    def name(self) -> str:
        return self.value.split(":")[0]

    @property
    def version(self) -> int:
        if ":" not in self.value:
            return 0
        return int(self.value.split(":")[1])


class IRLabel(IROperand):
    """
    IRLabel represents a label in IR. A label is a string that starts with a %.
    """

    # is_symbol is used to indicate if the label came from upstream
    # (like a function name, try to preserve it in optimization passes)
    is_symbol: bool = False
    value: str

    def __init__(self, value: str, is_symbol: bool = False) -> None:
        assert isinstance(value, str), "value must be an str"
        self.value = value
        self.is_symbol = is_symbol

    def __eq__(self, other):
        # no need for is_symbol to participate in equality
        return super().__eq__(other)

    def __hash__(self):
        # __hash__ is required when __eq__ is overridden --
        # https://docs.python.org/3/reference/datamodel.html#object.__hash__
        return super().__hash__()


class IRInstruction:
    """
    IRInstruction represents an instruction in IR. Each instruction has an opcode,
    operands, and return value. For example, the following IR instruction:
        %1 = add %0, 1
    has opcode "add", operands ["%0", "1"], and return value "%1".

    Convention: the rightmost value is the top of the stack.
    """

    opcode: str
    operands: list[IROperand]
    output: Optional[IROperand]
    # set of live variables at this instruction
    liveness: OrderedSet[IRVariable]
    dup_requirements: OrderedSet[IRVariable]
    parent: "IRBasicBlock"
    fence_id: int
    annotation: Optional[str]
    ast_source: Optional[IRnode]
    error_msg: Optional[str]

    def __init__(
        self,
        opcode: str,
        operands: list[IROperand] | Iterator[IROperand],
        output: Optional[IROperand] = None,
    ):
        assert isinstance(opcode, str), "opcode must be an str"
        assert isinstance(operands, list | Iterator), "operands must be a list"
        self.opcode = opcode
        self.operands = list(operands)  # in case we get an iterator
        self.output = output
        self.liveness = OrderedSet()
        self.dup_requirements = OrderedSet()
        self.fence_id = -1
        self.annotation = None
        self.ast_source = None
        self.error_msg = None

    @property
    def volatile(self) -> bool:
        return self.opcode in VOLATILE_INSTRUCTIONS

    def get_label_operands(self) -> Iterator[IRLabel]:
        """
        Get all labels in instruction.
        """
        return (op for op in self.operands if isinstance(op, IRLabel))

    def get_non_label_operands(self) -> Iterator[IROperand]:
        """
        Get input operands for instruction which are not labels
        """
        return (op for op in self.operands if not isinstance(op, IRLabel))

    def get_inputs(self) -> Iterator[IRVariable]:
        """
        Get all input operands for instruction.
        """
        return (op for op in self.operands if isinstance(op, IRVariable))

    def get_outputs(self) -> list[IROperand]:
        """
        Get the output item for an instruction.
        (Currently all instructions output at most one item, but write
        it as a list to be generic for the future)
        """
        return [self.output] if self.output else []

    def replace_operands(self, replacements: dict) -> None:
        """
        Update operands with replacements.
        replacements are represented using a dict: "key" is replaced by "value".
        """
        for i, operand in enumerate(self.operands):
            if operand in replacements:
                self.operands[i] = replacements[operand]

    def replace_label_operands(self, replacements: dict) -> None:
        """
        Update label operands with replacements.
        replacements are represented using a dict: "key" is replaced by "value".
        """
        replacements = {k.value: v for k, v in replacements.items()}
        for i, operand in enumerate(self.operands):
            if isinstance(operand, IRLabel) and operand.value in replacements:
                self.operands[i] = replacements[operand.value]

    @property
    def phi_operands(self) -> Iterator[tuple[IRLabel, IROperand]]:
        """
        Get phi operands for instruction.
        """
        assert self.opcode == "phi", "instruction must be a phi"
        for i in range(0, len(self.operands), 2):
            label = self.operands[i]
            var = self.operands[i + 1]
            assert isinstance(label, IRLabel), "phi operand must be a label"
            assert isinstance(
                var, (IRVariable, IRLiteral)
            ), "phi operand must be a variable or literal"
            yield label, var

    def remove_phi_operand(self, label: IRLabel) -> None:
        """
        Remove a phi operand from the instruction.
        """
        assert self.opcode == "phi", "instruction must be a phi"
        for i in range(0, len(self.operands), 2):
            if self.operands[i] == label:
                del self.operands[i : i + 2]
                return

    def get_ast_source(self) -> Optional[IRnode]:
        if self.ast_source:
            return self.ast_source
        idx = self.parent.instructions.index(self)
        for inst in reversed(self.parent.instructions[:idx]):
            if inst.ast_source:
                return inst.ast_source
        return self.parent.parent.ast_source

    def __repr__(self) -> str:
        s = ""
        if self.output:
            s += f"{self.output} = "
        opcode = f"{self.opcode} " if self.opcode != "store" else ""
        s += opcode
        operands = self.operands
        if opcode not in ["jmp", "jnz", "invoke"]:
            operands = reversed(operands)  # type: ignore
        s += ", ".join(
            [(f"label %{op}" if isinstance(op, IRLabel) else str(op)) for op in operands]
        )

        if self.annotation:
            s += f" <{self.annotation}>"

        if self.liveness:
            return f"{s: <30} # {self.liveness}"

        return s


def _ir_operand_from_value(val: Any) -> IROperand:
    if isinstance(val, IROperand):
        return val

    assert isinstance(val, int), val
    return IRLiteral(val)


class IRBasicBlock:
    """
    IRBasicBlock represents a basic block in IR. Each basic block has a label and
    a list of instructions, while belonging to a function.

    The following IR code:
        %1 = add %0, 1
        %2 = mul %1, 2
    is represented as:
        bb = IRBasicBlock("bb", function)
        r1 = bb.append_instruction("add", "%0", "1")
        r2 = bb.append_instruction("mul", r1, "2")

    The label of a basic block is used to refer to it from other basic blocks
    in order to branch to it.

    The parent of a basic block is the function it belongs to.

    The instructions of a basic block are executed sequentially, and the last
    instruction of a basic block is always a terminator instruction, which is
    used to branch to other basic blocks.
    """

    label: IRLabel
    parent: "IRFunction"
    instructions: list[IRInstruction]
    # basic blocks which can jump to this basic block
    cfg_in: OrderedSet["IRBasicBlock"]
    # basic blocks which this basic block can jump to
    cfg_out: OrderedSet["IRBasicBlock"]
    # stack items which this basic block produces
    out_vars: OrderedSet[IRVariable]

    reachable: OrderedSet["IRBasicBlock"]
    is_reachable: bool = False

    def __init__(self, label: IRLabel, parent: "IRFunction") -> None:
        assert isinstance(label, IRLabel), "label must be an IRLabel"
        self.label = label
        self.parent = parent
        self.instructions = []
        self.cfg_in = OrderedSet()
        self.cfg_out = OrderedSet()
        self.out_vars = OrderedSet()
        self.reachable = OrderedSet()
        self.is_reachable = False

    def add_cfg_in(self, bb: "IRBasicBlock") -> None:
        self.cfg_in.add(bb)

    def remove_cfg_in(self, bb: "IRBasicBlock") -> None:
        assert bb in self.cfg_in
        self.cfg_in.remove(bb)

    def add_cfg_out(self, bb: "IRBasicBlock") -> None:
        # malformed: jnz condition label1 label1
        # (we could handle but it makes a lot of code easier
        # if we have this assumption)
        self.cfg_out.add(bb)

    def remove_cfg_out(self, bb: "IRBasicBlock") -> None:
        assert bb in self.cfg_out
        self.cfg_out.remove(bb)

    def append_instruction(
        self, opcode: str, *args: Union[IROperand, int], ret: IRVariable = None
    ) -> Optional[IRVariable]:
        """
        Append an instruction to the basic block

        Returns the output variable if the instruction supports one
        """
        assert not self.is_terminated, self

        if ret is None:
            ret = self.parent.get_next_variable() if opcode not in NO_OUTPUT_INSTRUCTIONS else None

        # Wrap raw integers in IRLiterals
        inst_args = [_ir_operand_from_value(arg) for arg in args]

        inst = IRInstruction(opcode, inst_args, ret)
        inst.parent = self
        inst.ast_source = self.parent.ast_source
        inst.error_msg = self.parent.error_msg
        self.instructions.append(inst)
        return ret

    def append_invoke_instruction(
        self, args: list[IROperand | int], returns: bool
    ) -> Optional[IRVariable]:
        """
        Append an invoke to the basic block
        """
        assert not self.is_terminated, self
        ret = None
        if returns:
            ret = self.parent.get_next_variable()

        # Wrap raw integers in IRLiterals
        inst_args = [_ir_operand_from_value(arg) for arg in args]

        assert isinstance(inst_args[0], IRLabel), "Invoked non label"

        inst = IRInstruction("invoke", inst_args, ret)
        inst.parent = self
        inst.ast_source = self.parent.ast_source
        inst.error_msg = self.parent.error_msg
        self.instructions.append(inst)
        return ret

    def insert_instruction(self, instruction: IRInstruction, index: Optional[int] = None) -> None:
        assert isinstance(instruction, IRInstruction), "instruction must be an IRInstruction"

        if index is None:
            assert not self.is_terminated, self
            index = len(self.instructions)
        instruction.parent = self
        instruction.ast_source = self.parent.ast_source
        instruction.error_msg = self.parent.error_msg
        self.instructions.insert(index, instruction)

    def remove_instruction(self, instruction: IRInstruction) -> None:
        assert isinstance(instruction, IRInstruction), "instruction must be an IRInstruction"
        self.instructions.remove(instruction)

    def clear_instructions(self) -> None:
        self.instructions = []

    def replace_operands(self, replacements: dict) -> None:
        """
        Update operands with replacements.
        """
        for instruction in self.instructions:
            instruction.replace_operands(replacements)

    def get_assignments(self):
        """
        Get all assignments in basic block.
        """
        return [inst.output for inst in self.instructions if inst.output]

    def get_uses(self) -> dict[IRVariable, OrderedSet[IRInstruction]]:
        uses: dict[IRVariable, OrderedSet[IRInstruction]] = {}
        for inst in self.instructions:
            for op in inst.get_inputs():
                if op not in uses:
                    uses[op] = OrderedSet()
                uses[op].add(inst)
        return uses

    @property
    def is_empty(self) -> bool:
        """
        Check if the basic block is empty, i.e. it has no instructions.
        """
        return len(self.instructions) == 0

    @property
    def is_terminated(self) -> bool:
        """
        Check if the basic block is terminal, i.e. the last instruction is a terminator.
        """
        # it's ok to return False here, since we use this to check
        # if we can/need to append instructions to the basic block.
        if len(self.instructions) == 0:
            return False
        return self.instructions[-1].opcode in BB_TERMINATORS

    @property
    def is_terminal(self) -> bool:
        """
        Check if the basic block is terminal.
        """
        return len(self.cfg_out) == 0

    @property
    def in_vars(self) -> OrderedSet[IRVariable]:
        for inst in self.instructions:
            if inst.opcode != "phi":
                return inst.liveness
        return OrderedSet()

    def copy(self):
        bb = IRBasicBlock(self.label, self.parent)
        bb.instructions = self.instructions.copy()
        bb.cfg_in = self.cfg_in.copy()
        bb.cfg_out = self.cfg_out.copy()
        bb.out_vars = self.out_vars.copy()
        return bb

    def __repr__(self) -> str:
        s = (
            f"{repr(self.label)}:  IN={[bb.label for bb in self.cfg_in]}"
            f" OUT={[bb.label for bb in self.cfg_out]} => {self.out_vars}\n"
        )
        for instruction in self.instructions:
            s += f"    {str(instruction).strip()}\n"
        return s



import pytest

from vyper import compiler
from vyper.exceptions import (
    FlagDeclarationException,
    InvalidOperation,
    NamespaceCollision,
    StructureException,
    TypeMismatch,
    UnknownAttribute,
)

fail_list = [
    (
        """
event Action:
    pass

flag Action:
    BUY
    SELL
    """,
        NamespaceCollision,
    ),
    (
        """
flag Action:
    pass
    """,
        FlagDeclarationException,
    ),
    (
        """
flag Action:
    BUY
    BUY
    """,
        FlagDeclarationException,
    ),
    ("flag Foo:\n" + "\n".join([f"    member{i}" for i in range(257)]), FlagDeclarationException),
    (
        """
flag Roles:
    USER
    STAFF
    ADMIN

@external
def foo(x: Roles) -> bool:
    return x in [Roles.USER, Roles.ADMIN]
    """,
        TypeMismatch,
    ),
    (
        """
flag Roles:
    USER
    STAFF
    ADMIN

@external
def foo(x: Roles) -> Roles:
    return x.USER  # can't dereference on flag instance
    """,
        StructureException,
    ),
    (
        """
flag Roles:
    USER
    STAFF
    ADMIN

@external
def foo(x: Roles) -> bool:
    return x >= Roles.STAFF
    """,
        InvalidOperation,
    ),
    (
        """
flag Functions:
    def foo():nonpayable
    """,
        FlagDeclarationException,
    ),
    (
        """
flag Numbers:
    a:constant(uint256) = a
    """,
        FlagDeclarationException,
    ),
    (
        """
flag Numbers:
    12
    """,
        FlagDeclarationException,
    ),
    (
        """
flag Roles:
    ADMIN
    USER

@external
def foo() -> Roles:
    return Roles.GUEST
    """,
        UnknownAttribute,
    ),
    (
        """
flag A:
    a
flag B:
    a
    b

@internal
def foo():
    a: A = B.b
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_fail_cases(bad_code):
    with pytest.raises(bad_code[1]):
        compiler.compile_code(bad_code[0])


valid_list = [
    """
flag Action:
    BUY
    SELL
    """,
    """
flag Action:
    BUY
    SELL
@external
def run() -> Action:
    return Action.BUY
    """,
    """
flag Action:
    BUY
    SELL

struct Order:
    action: Action
    amount: uint256

@external
def run() -> Order:
    return Order(
        action=Action.BUY,
        amount=10**18
        )
    """,
    "flag Foo:\n" + "\n".join([f"    member{i}" for i in range(256)]),
    """
a: constant(uint256) = 1

flag A:
    a
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_flag_success(good_code):
    assert compiler.compile_code(good_code) is not None
